# Copyright (C) 2015-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import abc
import functools
import importlib
import inspect
import logging
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
import warnings

from deprecated import deprecated
from flask import Flask, Request, Response, abort, request
import requests
import sentry_sdk
from werkzeug.exceptions import HTTPException

from .negotiation import Formatter as FormatterBase
from .negotiation import Negotiator as NegotiatorBase
from .negotiation import negotiate as _negotiate
from .serializers import (
    decode_response,
    encode_data_client,
    exception_to_dict,
    json_dumps,
    json_loads,
    msgpack_dumps,
    msgpack_loads,
)

logger = logging.getLogger(__name__)


# support for content negotiation


class Negotiator(NegotiatorBase):
    def best_mimetype(self):
        return request.accept_mimetypes.best_match(
            self.accept_mimetypes, "application/json"
        )

    def _abort(self, status_code, err=None):
        return abort(status_code, err)


def negotiate(formatter_cls, *args, **kwargs):
    return _negotiate(Negotiator, formatter_cls, *args, **kwargs)


class Formatter(FormatterBase):
    def _make_response(self, body, content_type):
        return Response(body, content_type=content_type)

    def configure(self, extra_encoders=None):
        self.extra_encoders = extra_encoders


class JSONFormatter(Formatter):
    format = "json"
    mimetypes = ["application/json"]

    def render(self, obj):
        return json_dumps(obj, extra_encoders=self.extra_encoders)


class MsgpackFormatter(Formatter):
    format = "msgpack"
    mimetypes = ["application/x-msgpack"]

    def render(self, obj):
        return msgpack_dumps(obj, extra_encoders=self.extra_encoders)


# base API classes


class RemoteException(Exception):
    """raised when remote returned an out-of-band failure notification, e.g., as a
    HTTP status code or serialized exception

    Attributes:
        response: HTTP response corresponding to the failure

    """

    def __init__(
        self,
        payload: Optional[Any] = None,
        response: Optional[requests.Response] = None,
    ):
        if payload is not None:
            super().__init__(payload)
        else:
            super().__init__()
        self.response = response

    def __str__(self):
        if (
            self.args
            and isinstance(self.args[0], dict)
            and "type" in self.args[0]
            and "args" in self.args[0]
        ):
            return (
                f"<RemoteException {self.response.status_code} "
                f'{self.args[0]["type"]}: {self.args[0]["args"]}>'
            )
        else:
            return super().__str__()


class TransientRemoteException(RemoteException):
    """Subclass of RemoteException representing errors which are expected
    to be temporary.
    """


F = TypeVar("F", bound=Callable)


def remote_api_endpoint(path: str, method: str = "POST") -> Callable[[F], F]:
    def dec(f: F) -> F:
        f._endpoint_path = path  # type: ignore
        f._method = method  # type: ignore
        return f

    return dec


class APIError(Exception):
    """API Error"""

    def __str__(self):
        return "An unexpected error occurred in the backend: {}".format(self.args)


class MetaRPCClient(type):
    """Metaclass for RPCClient, which adds a method for each endpoint
    of the database it is designed to access.

    See for example :class:`swh.indexer.storage.api.client.RemoteStorage`"""

    def __new__(cls, name, bases, attributes):
        # For each method wrapped with @remote_api_endpoint in an API backend
        # (eg. :class:`swh.indexer.storage.IndexerStorage`), add a new
        # method in RemoteStorage, with the same documentation.
        #
        # Note that, despite the usage of decorator magic (eg. functools.wrap),
        # this never actually calls an IndexerStorage method.
        backend_class = attributes.get("backend_class", None)
        for base in bases:
            if backend_class is not None:
                break
            backend_class = getattr(base, "backend_class", None)
        if backend_class:
            for meth_name, meth in backend_class.__dict__.items():
                if hasattr(meth, "_endpoint_path"):
                    cls.__add_endpoint(meth_name, meth, attributes)
        return super().__new__(cls, name, bases, attributes)

    @staticmethod
    def __add_endpoint(meth_name, meth, attributes):
        wrapped_meth = inspect.unwrap(meth)

        @functools.wraps(meth)  # Copy signature and doc
        def meth_(*args, **kwargs):
            # Match arguments and parameters
            post_data = inspect.getcallargs(wrapped_meth, *args, **kwargs)

            # Remove arguments that should not be passed
            self = post_data.pop("self")
            post_data.pop("cur", None)
            post_data.pop("db", None)

            # Send the request.
            return self._post(meth._endpoint_path, post_data)

        if meth_name not in attributes:
            attributes[meth_name] = meth_


class RPCClient(metaclass=MetaRPCClient):
    """Proxy to an internal SWH RPC.

    Arguments:
      url: base url for the RPC endpoints
      timeout: request timeout. Can be a pair of floats, used to distinguish
        between connection and read timeouts.
      chunk_size: used for iteration on chunked responses
      max_retries: number of automatic retries issued for requests
      pool_connections: number of connections instantiated in the default connection pool
      pool_maxsize: maximum size of the connection pool
      adapter_kwargs: extra keyword arguments to pass to the
        :class:`requests.adapters.HTTPAdapter`
      api_exception: The exception class to raise in case of communication error
        with the server.
      reraise_exceptions: On server errors, if any of the exception classes in
        this list has the same name as the error name, then the exception will
        be instantiated and raised instead of a generic RemoteException.

    """

    backend_class: ClassVar[Optional[type]] = None
    """For each method of `backend_class` decorated with
    :func:`remote_api_endpoint`, a method with the same prototype and
    docstring will be added to this class. Calls to this new method will
    be translated into HTTP requests to a remote server.

    This backend class will never be instantiated, it only serves as
    a template."""

    api_exception: Type[Exception] = APIError
    """The exception class to raise in case of communication error with
    the server."""

    reraise_exceptions: List[Type[Exception]] = []
    """On server errors, if any of the exception classes in this list
    has the same name as the error name, then the exception will
    be instantiated and raised instead of a generic RemoteException."""

    extra_type_encoders: List[Tuple[type, str, Callable]] = []
    """Value of `extra_encoders` passed to `json_dumps` or `msgpack_dumps`
    to be able to serialize more object types."""
    extra_type_decoders: Dict[str, Callable] = {}
    """Value of `extra_decoders` passed to `json_loads` or `msgpack_loads`
    to be able to deserialize more object types."""

    def __init__(
        self,
        url: str,
        timeout: Union[None, Tuple[float, float], List[float], float] = None,
        chunk_size: int = 4096,
        max_retries: int = 3,
        pool_connections: int = 20,
        pool_maxsize: int = 100,
        adapter_kwargs: Optional[Dict[str, Any]] = None,
        api_exception: Optional[Type[Exception]] = None,
        reraise_exceptions: Optional[List[Type[Exception]]] = None,
        **kwargs,
    ):
        if api_exception:
            self.api_exception = api_exception
        if reraise_exceptions:
            self.reraise_exceptions = reraise_exceptions
        base_url = url if url.endswith("/") else url + "/"
        self.url = base_url

        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            max_retries=max_retries,
            pool_connections=pool_connections,
            pool_maxsize=pool_maxsize,
            **(adapter_kwargs or {}),
        )
        self.session.mount(self.url, adapter)

        if isinstance(timeout, list):
            if len(timeout) != 2:
                raise ValueError(
                    "timeout must be a pair of (connect, read) floats, not %r" % timeout
                )
            timeout = (timeout[0], timeout[1])
        self.timeout = timeout

        self.chunk_size = chunk_size

    def _url(self, endpoint):
        return "%s%s" % (self.url, endpoint)

    def raw_verb(self, verb, endpoint, **opts):
        if "chunk_size" in opts:
            # if the chunk_size argument has been passed, consider the user
            # also wants stream=True, otherwise, what's the point.
            opts["stream"] = True
        if self.timeout and "timeout" not in opts:
            opts["timeout"] = self.timeout
        try:
            return getattr(self.session, verb)(self._url(endpoint), **opts)
        except requests.exceptions.ConnectionError as e:
            raise self.api_exception(e)

    def _post(self, endpoint, data, **opts):
        if isinstance(data, (abc.Iterator, abc.Generator)):
            data = (self._encode_data(x) for x in data)
        else:
            data = self._encode_data(data)
        chunk_size = opts.pop("chunk_size", self.chunk_size)
        response = self.raw_verb(
            "post",
            endpoint,
            data=data,
            headers={
                "content-type": "application/x-msgpack",
                "accept": "application/x-msgpack",
            },
            **opts,
        )
        if opts.get("stream") or response.headers.get("transfer-encoding") == "chunked":
            self.raise_for_status(response)
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    def _encode_data(self, data):
        return encode_data_client(data, extra_encoders=self.extra_type_encoders)

    _post_stream = _post

    @deprecated(version="2.1.0", reason="Use _post instead")
    def post(self, *args, **kwargs):
        return self._post(*args, **kwargs)

    @deprecated(version="2.1.0", reason="Use _post_stream instead")
    def post_stream(self, *args, **kwargs):
        return self._post_stream(*args, **kwargs)

    def _get(self, endpoint, **opts):
        chunk_size = opts.pop("chunk_size", self.chunk_size)
        response = self.raw_verb(
            "get", endpoint, headers={"accept": "application/x-msgpack"}, **opts
        )
        if opts.get("stream") or response.headers.get("transfer-encoding") == "chunked":
            self.raise_for_status(response)
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    def _get_stream(self, endpoint, **opts):
        return self._get(endpoint, stream=True, **opts)

    @deprecated(version="2.1.0", reason="Use _get instead")
    def get(self, *args, **kwargs):
        return self._get(*args, **kwargs)

    @deprecated(version="2.1.0", reason="Use _get_stream instead")
    def get_stream(self, *args, **kwargs):
        return self._get_stream(*args, **kwargs)

    def raise_for_status(self, response) -> None:
        """check response HTTP status code and raise an exception if it denotes an
        error; do nothing otherwise

        """
        status_code = response.status_code
        status_class = response.status_code // 100

        if status_code == 404:
            raise RemoteException(payload="404 not found", response=response)

        exception = None

        if status_class == 4:
            exc_data = self._decode_response(response, check_status=False)
            if isinstance(exc_data, dict):
                for exc_type in self.reraise_exceptions:
                    if exc_type.__name__ == exc_data["type"]:
                        exception = exc_type(*exc_data["args"])
                        break
                else:
                    exception = RemoteException(payload=exc_data, response=response)
            else:
                # Typically, because the error is from a reverse proxy not aware of this
                # RPC protocol, so response's content-type is text/html
                exception = APIError(exc_data, response)

        elif status_class == 5:
            cls: Type[RemoteException]
            if status_code in (502, 503):
                # This isn't a generic HTTP client and we know the server does
                # not support the Retry-After header, so we do not implement
                # it here either.
                cls = TransientRemoteException
            else:
                cls = RemoteException
            exc_data = self._decode_response(response, check_status=False)
            exception = cls(payload=exc_data, response=response)

        if exception:
            raise exception from None

        if status_class != 2:
            raise RemoteException(
                payload=f"API HTTP error: {status_code} {response.content}",
                response=response,
            )

    def _decode_response(self, response, check_status=True):
        if check_status:
            self.raise_for_status(response)
        return decode_response(response, extra_decoders=self.extra_type_decoders)

    def __repr__(self):
        return "<{} url={}>".format(self.__class__.__name__, self.url)


class BytesRequest(Request):
    """Request with proper escaping of arbitrary byte sequences."""

    encoding = "utf-8"
    encoding_errors = "surrogateescape"


ENCODERS: Dict[str, Callable[[Any], Union[bytes, str]]] = {
    "application/x-msgpack": msgpack_dumps,
    "application/json": json_dumps,
}


def encode_data_server(
    data, content_type="application/x-msgpack", extra_type_encoders=None
):
    encoded_data = ENCODERS[content_type](data, extra_encoders=extra_type_encoders)
    return Response(
        encoded_data,
        mimetype=content_type,
    )


def decode_request(request, extra_decoders=None):
    content_type = request.mimetype
    data = request.get_data()
    if not data:
        return {}

    if content_type == "application/x-msgpack":
        r = msgpack_loads(data, extra_decoders=extra_decoders)
    elif content_type == "application/json":
        # XXX this .decode() is needed for py35.
        # Should not be needed any more with py37
        r = json_loads(data.decode("utf-8"), extra_decoders=extra_decoders)
    else:
        raise ValueError("Wrong content type `%s` for API request" % content_type)

    return r


def error_handler(
    exception: BaseException, encoder=encode_data_server, status_code: int = 500
):
    """Error handler to be registered using flask's error-handling decorator
    ``app.errorhandler``.

    This is used for exceptions that are expected in the normal execution flow of the
    RPC-ed API, in which case the status code should be set to a value in the 4xx range,
    as well as for exceptions that are unexpected (generally, a bare
    :class:`Exception`), and for which the status code should be kept in the 5xx class.

    This function only captures exceptions as sentry errors if the status code is in the
    5xx range and not 502/503/504, as "expected exceptions" in the 4xx range are more,
    likely to be handled on the client side; and 502/503/504 are "transient" exceptions
    that should be resolved with client retries.

    """
    status_class = status_code // 100
    if status_class == 5 and status_code not in (502, 503, 504):
        logger.exception(exception)
        sentry_sdk.capture_exception(exception)

    response = encoder(exception_to_dict(exception))
    if isinstance(exception, HTTPException):
        response.status_code = exception.code
    else:
        response.status_code = status_code
    return response


class RPCServerApp(Flask):
    """For each endpoint of the given `backend_class`, tells app.route to call
    a function that decodes the request and sends it to the backend object
    provided by the factory.

    :param Any backend_class:
        The class of the backend, which will be analyzed to look
        for API endpoints.
    :param Optional[Callable[[], backend_class]] backend_factory:
        A function with no argument that returns an instance of
        `backend_class`. If unset, defaults to calling `backend_class`
        constructor directly.

    For each method 'do_x()' of the ``backend_factory``, subclasses may implement
    two methods: ``pre_do_x(self, kw)`` and ``post_do_x(self, ret, kw)`` that will
    be called respectively before and after ``do_x(**kw)``. ``kw`` is the dict
    of request parameters, and ``ret`` is the return value of ``do_x(**kw)``.
    """

    request_class = BytesRequest

    extra_type_encoders: List[Tuple[type, str, Callable]] = []
    """Value of `extra_encoders` passed to `json_dumps` or `msgpack_dumps`
    to be able to serialize more object types."""
    extra_type_decoders: Dict[str, Callable] = {}
    """Value of `extra_decoders` passed to `json_loads` or `msgpack_loads`
    to be able to deserialize more object types."""

    method_decorators: List[Callable[[Callable], Callable]] = []
    """List of decorators to all methods generated from the ``backend_class``."""

    exception_status_codes: List[Tuple[Union[Type[BaseException], str], int]] = [
        # Default to "Internal Server Error" for most exceptions:
        (Exception, 500),
        # These errors are noisy, and are better logged on the caller's side after
        # it retried a few times:
        ("psycopg2.errors.OperationalError", 503),
        # Subclass of OperationalError; but it is unlikely to be solved after retries
        # (short of getting more cache hits) because this is usually caused by the query
        # size instead of a transient failure
        ("psycopg2.errors.QueryCanceled", 500),
        # Often a transient error because of connectivity issue with, or restart of,
        # the Kafka brokers:
        ("swh.journal.writer.kafka.KafkaDeliveryError", 503),
    ]
    """Pairs of ``(exception, status_code)`` where ``exception`` is either an
    exception class or a a dotted exception name to be imported (and ignored if import
    fails) and ``status_code`` is the HTTP code that should be returned when an instance
    of this exception is raised.

    If a raised exception is an instance of a subclass of two classes defined here,
    the most specific class wins, according Flask's MRO-based resolution."""

    def __init__(self, *args, backend_class=None, backend_factory=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_backend_class(backend_class, backend_factory)
        self._register_error_handlers()

    def _register_error_handlers(self):
        for exception, status_code in self.exception_status_codes:
            if isinstance(exception, str):
                (module_path, class_name) = exception.rsplit(".", 1)
                try:
                    module = importlib.import_module(module_path, package=__package__)
                except ImportError as e:
                    logger.debug("Could not import %s: %r", exception, e)
                    continue
                exception = getattr(module, class_name)

            self.register_error_handler(
                exception, functools.partial(error_handler, status_code=status_code)
            )

    def add_backend_class(self, backend_class=None, backend_factory=None):
        if backend_class is None and backend_factory is not None:
            raise ValueError(
                "backend_factory should only be provided if backend_class is"
            )

        if backend_class is not None:
            backend_factory = backend_factory or backend_class
            for meth_name, meth in backend_class.__dict__.items():
                if hasattr(meth, "_endpoint_path"):
                    self.__add_endpoint(meth_name, meth, backend_factory)

    def __add_endpoint(self, meth_name, meth, backend_factory):
        from flask import request

        @negotiate(MsgpackFormatter, extra_encoders=self.extra_type_encoders)
        @negotiate(JSONFormatter, extra_encoders=self.extra_type_encoders)
        @functools.wraps(meth)  # Copy signature and doc
        def f():
            # Call the actual code
            pre_hook = getattr(self, f"pre_{meth_name}", None)
            post_hook = getattr(self, f"post_{meth_name}", None)
            obj_meth = getattr(backend_factory(), meth_name)
            kw = decode_request(request, extra_decoders=self.extra_type_decoders)

            if pre_hook is not None:
                pre_hook(kw)

            ret = obj_meth(**kw)

            if post_hook is not None:
                post_hook(ret, kw)

            return ret

        for decorator in self.method_decorators:
            f = decorator(f)

        self.route("/" + meth._endpoint_path, methods=["POST"])(f)

    def setup_psycopg2_errorhandlers(self) -> None:
        """Deprecated method; error handlers are now setup in the constructor."""
        warnings.warn(
            "setup_psycopg2_errorhandlers has no effect; error handlers are now setup "
            "by the constructor.",
            DeprecationWarning,
        )



# Copyright (C) 2016-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import codecs
from contextlib import contextmanager
import itertools
import os
import re
from typing import Iterable, Tuple, TypeVar


@contextmanager
def cwd(path):
    """Contextually change the working directory to do thy bidding.
    Then gets back to the original location.

    """
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def grouper(iterable, n):
    """
    Collect data into fixed-length size iterables. The last block might
    contain less elements as it will hold only the remaining number
    of elements.

    The invariant here is that the number of elements in the input
    iterable and the sum of the number of elements of all iterables
    generated from this function should be equal.

    If ``iterable`` is an iterable of bytes or strings that you need to join
    later, then :func:`iter_chunks`` is preferable, as it avoids this join
    by slicing directly.

    Args:
        iterable (Iterable): an iterable
        n (int): size of block to slice the iterable into

    Yields:
        fixed-length blocks as iterables. As mentioned, the last
        iterable might be less populated.

    """
    args = [iter(iterable)] * n
    stop_value = object()
    for _data in itertools.zip_longest(*args, fillvalue=stop_value):
        yield (d for d in _data if d is not stop_value)


TStr = TypeVar("TStr", bytes, str)


def iter_chunks(
    iterable: Iterable[TStr], chunk_size: int, *, remainder: bool = False
) -> Iterable[TStr]:
    """
    Reads ``bytes`` objects (resp. ``str`` objects) from the ``iterable``,
    and yields them as chunks of exactly ``chunk_size`` bytes (resp. characters).

    ``iterable`` is typically obtained by repeatedly calling a method like
    :meth:`io.RawIOBase.read`; which does only guarantees an upper bound on the size;
    whereas this function returns chunks of exactly the size.

    Args:
        iterable: the input data
        chunk_size: the exact size of chunks to return
        remainder: if True, a last chunk with size strictly smaller than ``chunk_size``
          may be returned, if the data stream from the ``iterable`` had a length that
          is not a multiple of ``chunk_size``
    """
    buf = None
    iterator = iter(iterable)
    while True:
        assert buf is None or len(buf) < chunk_size
        try:
            new_data = next(iterator)
        except StopIteration:
            if remainder and buf:
                yield buf  # may be shorter than ``chunk_size``
            return

        if buf:
            buf += new_data
        else:
            # spares a copy
            buf = new_data

        new_buf = None
        for i in range(0, len(buf), chunk_size):
            chunk = buf[i : i + chunk_size]
            if len(chunk) == chunk_size:
                yield chunk
            else:
                assert not new_buf
                new_buf = chunk
        buf = new_buf


def backslashescape_errors(exception):
    if isinstance(exception, UnicodeDecodeError):
        bad_data = exception.object[exception.start : exception.end]
        escaped = "".join(r"\x%02x" % x for x in bad_data)
        return escaped, exception.end

    return codecs.backslashreplace_errors(exception)


codecs.register_error("backslashescape", backslashescape_errors)


def encode_with_unescape(value):
    """Encode an unicode string containing \\x<hex> backslash escapes"""
    slices = []
    start = 0
    odd_backslashes = False
    i = 0
    while i < len(value):
        if value[i] == "\\":
            odd_backslashes = not odd_backslashes
        else:
            if odd_backslashes:
                if value[i] != "x":
                    raise ValueError(
                        "invalid escape for %r at position %d" % (value, i - 1)
                    )
                slices.append(
                    value[start : i - 1].replace("\\\\", "\\").encode("utf-8")
                )
                slices.append(bytes.fromhex(value[i + 1 : i + 3]))

                odd_backslashes = False
                start = i = i + 3
                continue

        i += 1

    slices.append(value[start:i].replace("\\\\", "\\").encode("utf-8"))

    return b"".join(slices)


def decode_with_escape(value):
    """Decode a bytestring as utf-8, escaping the bytes of invalid utf-8 sequences
    as \\x<hex value>. We also escape NUL bytes as they are invalid in JSON
    strings.
    """
    # escape backslashes
    value = value.replace(b"\\", b"\\\\")
    value = value.replace(b"\x00", b"\\x00")
    return value.decode("utf-8", "backslashescape")


def commonname(path0, path1, as_str=False):
    """Compute the commonname between the path0 and path1."""
    return path1.split(path0)[1]


def numfile_sortkey(fname: str) -> Tuple[int, str]:
    """Simple function to sort filenames of the form:

      nnxxx.ext

    where nn is a number according to the numbers.

    Returns a tuple (order, remaining), where 'order' is the numeric (int)
    value extracted from the file name, and 'remaining' is the remaining part
    of the file name.

    Typically used to sort sql/nn-swh-xxx.sql files.

    Unmatched file names will return 999999 as order value.

    """
    m = re.match(r"(\d*)(.*)", fname)
    assert m is not None
    num, rem = m.groups()
    return (int(num) if num else 999999, rem)


def basename_sortkey(fname: str) -> Tuple[int, str]:
    "like numfile_sortkey but on basenames"
    return numfile_sortkey(os.path.basename(fname))



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from importlib.metadata import distribution
import logging
import textwrap
import traceback
from typing import List

import click
from click.testing import CliRunner
import pytest
import yaml

help_msg_snippets = (
    (
        "Usage",
        (
            "swh [OPTIONS] COMMAND [ARGS]...",
            "Command line interface for Software Heritage.",
        ),
    ),
    (
        "Options",
        (
            "-l, --log-level",
            "--log-config",
            "--sentry-dsn",
            "-h, --help",
        ),
    ),
)


def assert_result(result):
    if result.exception:
        assert result.exit_code == 0, (
            "Unexpected exception: "
            f"{''.join(traceback.format_tb(result.exc_info[2]))}"
            f"\noutput: {result.output}"
        )
    else:
        assert result.exit_code == 0, f"Unexpected output: {result.output}"


def get_section(cli_output: str, section: str) -> List[str]:
    """Get the given `section` of the `cli_output`"""
    result = []
    in_section = False
    for line in cli_output.splitlines():
        if not line:
            continue

        if in_section:
            if not line.startswith(" "):
                break
        else:
            if line.startswith(section):
                in_section = True

        if in_section:
            result.append(line)

    return result


def assert_section_contains(cli_output: str, section: str, snippet: str) -> bool:
    """Check that a given `section` of the `cli_output` contains the given `snippet`"""
    section_lines = get_section(cli_output, section)
    assert section_lines, "Section %s not found in output %r" % (section, cli_output)

    for line in section_lines:
        if snippet in line:
            return True
    else:
        assert False, "%r not found in section %r of output %r" % (
            snippet,
            section,
            cli_output,
        )


@pytest.fixture(autouse=True)
def reset_root_loglevel():
    root_level = logging.root.level
    yield
    logging.root.setLevel(root_level)


@pytest.fixture
def patched_dictconfig(mocker):
    yield mocker.patch("swh.core.logging.dictConfig", autospec=True)


def test_swh_help(swhmain):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["-h"])
    assert_result(result)
    for section, snippets in help_msg_snippets:
        for snippet in snippets:
            assert_section_contains(result.output, section, snippet)

    result = runner.invoke(swhmain, ["--help"])
    assert_result(result)
    for section, snippets in help_msg_snippets:
        for snippet in snippets:
            assert_section_contains(result.output, section, snippet)


def test_command(swhmain, mocker):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    result = runner.invoke(swhmain, ["test"])
    sentry_sdk_init.assert_called_once_with(
        dsn=None,
        debug=False,
        integrations=[],
        release=None,
        environment=None,
    )
    assert_result(result)
    assert result.output.strip() == "Hello SWH!"


def test_loglevel_default(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 20
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["test"])
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""


def test_loglevel_error(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 40
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["-l", "ERROR", "test"])
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""


def test_loglevel_debug(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 10
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["-l", "DEBUG", "test"])
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""


def test_sentry(swhmain, mocker):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    result = runner.invoke(swhmain, ["--sentry-dsn", "test_dsn", "test"])
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=False,
        integrations=[],
        release=None,
        environment=None,
    )


def test_sentry_debug(swhmain, mocker):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    result = runner.invoke(
        swhmain, ["--sentry-dsn", "test_dsn", "--sentry-debug", "test"]
    )
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=True,
        integrations=[],
        release=None,
        environment=None,
    )


def test_sentry_env(swhmain, mocker):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    env = {
        "SWH_SENTRY_DSN": "test_dsn",
        "SWH_SENTRY_DEBUG": "1",
    }
    result = runner.invoke(swhmain, ["test"], env=env, auto_envvar_prefix="SWH")
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=True,
        integrations=[],
        release=None,
        environment=None,
    )


def test_sentry_env_main_package(swhmain, mocker):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    env = {
        "SWH_SENTRY_DSN": "test_dsn",
        "SWH_MAIN_PACKAGE": "swh.core",
        "SWH_SENTRY_ENVIRONMENT": "tests",
    }
    result = runner.invoke(swhmain, ["test"], env=env, auto_envvar_prefix="SWH")
    assert_result(result)

    version = distribution("swh.core").version

    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=False,
        integrations=[],
        release="swh.core@" + version,
        environment="tests",
    )


@pytest.fixture
def log_config_path(tmp_path):
    log_config = textwrap.dedent(
        """\
    ---
    version: 1
    formatters:
      formatter:
        format: 'custom format:%(name)s:%(levelname)s:%(message)s'
    handlers:
      console:
        class: logging.StreamHandler
        stream: ext://sys.stdout
        formatter: formatter
        level: DEBUG
    root:
      level: DEBUG
      handlers:
        - console
    loggers:
      dontshowdebug:
        level: INFO
    """
    )

    (tmp_path / "log_config.yml").write_text(log_config)

    yield str(tmp_path / "log_config.yml")


def test_log_config(log_config_path, swhmain, patched_dictconfig):
    """Check that --log-config properly calls :func:`logging.config.dictConfig`."""

    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        patched_dictconfig.assert_called_once_with(
            yaml.safe_load(open(log_config_path, "r"))
        )

        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        [
            "--log-config",
            log_config_path,
            "test",
        ],
    )

    assert_result(result)
    assert result.output.strip() == "Hello SWH!"


def test_log_config_log_level_interaction(log_config_path, swhmain, patched_dictconfig):
    """Check that --log-config and --log-level work properly together, by calling
    :func:`logging.config.dictConfig` then setting the loglevel of the root logger."""

    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == logging.DEBUG
        patched_dictconfig.assert_called_once_with(
            yaml.safe_load(open(log_config_path, "r"))
        )

        click.echo("Hello SWH!")

    assert logging.root.level != logging.DEBUG

    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        [
            "--log-config",
            log_config_path,
            "--log-level",
            "DEBUG",
            "test",
        ],
    )

    assert_result(result)
    assert result.output.strip() == "Hello SWH!"


def test_multiple_log_level_behavior(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.getLevelName(logging.root.level) == "DEBUG"
        assert logging.getLevelName(logging.getLogger("dontshowdebug").level) == "INFO"
        return 0

    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        [
            "--log-level",
            "DEBUG",
            "--log-level",
            "dontshowdebug:INFO",
            "test",
        ],
    )

    assert_result(result)


def test_invalid_log_level(swhmain):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["--log-level", "broken:broken:DEBUG"])

    assert result.exit_code != 0
    assert "Invalid log level specification" in result.output

    runner = CliRunner()
    result = runner.invoke(swhmain, ["--log-level", "UNKNOWN"])

    assert result.exit_code != 0
    assert "Log level UNKNOWN unknown" in result.output


def test_aliased_command(swhmain):
    @swhmain.command(name="canonical-test")
    @click.pass_context
    def swhtest(ctx):
        "A test command."
        click.echo("Hello SWH!")

    swhmain.add_alias(swhtest, "othername")

    runner = CliRunner()

    # check we have only 'canonical-test' listed in the usage help msg
    result = runner.invoke(swhmain, ["-h"])
    assert_result(result)
    assert "canonical-test  A test command." in result.output
    assert "othername" not in result.output

    # check we can execute the cmd with 'canonical-test'
    result = runner.invoke(swhmain, ["canonical-test"])
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""

    # check we can also execute the cmd with the alias 'othername'
    result = runner.invoke(swhmain, ["othername"])
    assert_result(result)
    assert result.output.strip() == """Hello SWH!"""


def test_documentation(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        """Does a thing

        This needs the following config:

        \b
        * :ref:`cli-config-storage`
        * :ref:`cli-config-scheduler`

        and calls :mod:`swh.core.test.test_cli`
        """
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["test", "--help"])
    assert_result(result)
    assert result.output == textwrap.dedent(
        """\
        Usage: swh test [OPTIONS]

          Does a thing

          This needs the following config:

          * storage key: https://docs.softwareheritage.org/devel/configuration.html#cli-config-storage
          * scheduler key: https://docs.softwareheritage.org/devel/configuration.html#cli-config-scheduler

          and calls
          https://docs.softwareheritage.org/devel/apidoc/swh.core.test.test_cli.html

        Options:
          -h, --help  Show this message and exit.
        """  # noqa
    )



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from copy import deepcopy
from itertools import chain
import logging
import os
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple

from backports.entry_points_selectable import entry_points as get_entry_points
from deprecated import deprecated
import yaml

logger = logging.getLogger(__name__)


SWH_CONFIG_DIRECTORIES = [
    "~/.config/swh",
    "~/.swh",
    "/etc/softwareheritage",
]

SWH_GLOBAL_CONFIG = "global.yml"

SWH_DEFAULT_GLOBAL_CONFIG = {
    "max_content_size": ("int", 100 * 1024 * 1024),
}

SWH_CONFIG_EXTENSIONS = [
    ".yml",
]

# conversion per type
_map_convert_fn: Dict[str, Callable] = {
    "int": int,
    "bool": lambda x: x.lower() == "true",
    "list[str]": lambda x: [value.strip() for value in x.split(",")],
    "list[int]": lambda x: [int(value.strip()) for value in x.split(",")],
}

_map_check_fn: Dict[str, Callable] = {
    "int": lambda x: isinstance(x, int),
    "bool": lambda x: isinstance(x, bool),
    "list[str]": lambda x: (isinstance(x, list) and all(isinstance(y, str) for y in x)),
    "list[int]": lambda x: (isinstance(x, list) and all(isinstance(y, int) for y in x)),
}


def exists_accessible(filepath: str) -> bool:
    """Check whether a file exists, and is accessible.

    Returns:
        True if the file exists and is accessible
        False if the file does not exist

    Raises:
        PermissionError if the file cannot be read.
    """

    try:
        os.stat(filepath)
    except PermissionError:
        raise
    except (FileNotFoundError, NotADirectoryError):
        return False
    else:
        if os.access(filepath, os.R_OK):
            return True
        else:
            raise PermissionError("Permission denied: {filepath!r}")


def read_raw_config(base_config_path: str) -> Dict[str, Any]:
    """Read the raw config corresponding to base_config_path.

    Can read yml files.
    """
    yml_file = config_path(base_config_path)
    if yml_file is None:
        logging.error("Config file %s does not exist, ignoring it.", base_config_path)
        return {}
    else:
        logger.debug("Loading config file %s", yml_file)
        with open(yml_file) as f:
            return yaml.safe_load(f)


@deprecated(
    version="2.23.0",
    reason="pass config paths as-is to read_raw_config/read, and rely on click.Path",
)
def config_exists(path):
    """Check whether the given config exists"""
    path = config_path(path)
    return path is not None and exists_accessible(path)


@deprecated(version="2.23.0", reason="pass config paths as-is to read_raw_config/read")
def config_basepath(config_path: str) -> str:
    """Return the base path of a configuration file"""
    if config_path.endswith(".yml"):
        return config_path[:-4]

    return config_path


def config_path(config_path):
    """Check whether the given config exists"""
    if exists_accessible(config_path):
        return config_path
    for extension in SWH_CONFIG_EXTENSIONS:
        if exists_accessible(config_path + extension):
            logger.warning(
                "%s does not exist, using %s instead",
                config_path,
                config_path + extension,
            )
            return config_path + extension

    return None


def read(
    conf_file: Optional[str] = None,
    default_conf: Optional[Dict[str, Tuple[str, Any]]] = None,
) -> Dict[str, Any]:
    """Read the user's configuration file.

    Fill in the gap using `default_conf`.  `default_conf` is similar to this::

        DEFAULT_CONF = {
            'a': ('str', '/tmp/swh-loader-git/log'),
            'b': ('str', 'dbname=swhloadergit')
            'c': ('bool', true)
            'e': ('bool', None)
            'd': ('int', 10)
        }

    If conf_file is None, return the default config.

    """
    conf: Dict[str, Any] = {}

    if conf_file:
        base_config_path = os.path.expanduser(conf_file)
        conf = read_raw_config(base_config_path) or {}

    if not default_conf:
        return conf

    # remaining missing default configuration key are set
    # also type conversion is enforced for underneath layer
    for key, (nature_type, default_value) in default_conf.items():
        val = conf.get(key, None)
        if val is None:  # fallback to default value
            conf[key] = default_value
        elif not _map_check_fn.get(nature_type, lambda x: True)(val):
            # value present but not in the proper format, force type conversion
            conf[key] = _map_convert_fn.get(nature_type, lambda x: x)(val)

    return conf


def priority_read(
    conf_filenames: List[str], default_conf: Optional[Dict[str, Tuple[str, Any]]] = None
):
    """Try reading the configuration files from conf_filenames, in order,
    and return the configuration from the first one that exists.

    default_conf has the same specification as it does in read.
    """

    # Try all the files in order
    for filename in conf_filenames:
        full_filename = config_path(os.path.expanduser(filename))
        if full_filename is not None:
            return read(full_filename, default_conf)

    # Else, return the default configuration
    return read(None, default_conf)


def merge_default_configs(base_config, *other_configs):
    """Merge several default config dictionaries, from left to right"""
    full_config = base_config.copy()

    for config in other_configs:
        full_config.update(config)

    return full_config


def merge_configs(base: Optional[Dict[str, Any]], other: Optional[Dict[str, Any]]):
    """Merge two config dictionaries

    This does merge config dicts recursively, with the rules, for every value
    of the dicts (with 'val' not being a dict):

    - None + type -> type
    - type + None -> None
    - dict + dict -> dict (merged)
    - val + dict -> TypeError
    - dict + val -> TypeError
    - val + val -> val (other)

    for instance:

    >>> d1 = {
    ...   'key1': {
    ...     'skey1': 'value1',
    ...     'skey2': {'sskey1': 'value2'},
    ...   },
    ...   'key2': 'value3',
    ... }

    with

    >>> d2 = {
    ...   'key1': {
    ...     'skey1': 'value4',
    ...     'skey2': {'sskey2': 'value5'},
    ...   },
    ...   'key3': 'value6',
    ... }

    will give:

    >>> d3 = {
    ...   'key1': {
    ...     'skey1': 'value4',  # <-- note this
    ...     'skey2': {
    ...       'sskey1': 'value2',
    ...       'sskey2': 'value5',
    ...     },
    ...   },
    ...   'key2': 'value3',
    ...   'key3': 'value6',
    ... }
    >>> assert merge_configs(d1, d2) == d3

    Note that no type checking is done for anything but dicts.
    """
    if not isinstance(base, dict) or not isinstance(other, dict):
        raise TypeError("Cannot merge a %s with a %s" % (type(base), type(other)))

    output = {}
    for k in chain(base.keys(), other.keys()):
        if k in output:
            continue

        vb = base.get(k)
        vo = other.get(k)

        if isinstance(vo, dict):
            output[k] = merge_configs(vb is not None and vb or {}, vo)
        elif isinstance(vb, dict) and k in other and other[k] is not None:
            output[k] = merge_configs(vb, vo is not None and vo or {})
        elif k in other:
            output[k] = deepcopy(vo)
        else:
            output[k] = deepcopy(vb)

    return output


def swh_config_paths(base_filename: str) -> List[str]:
    """Return the Software Heritage specific configuration paths for the given
    filename."""

    return [os.path.join(dirname, base_filename) for dirname in SWH_CONFIG_DIRECTORIES]


def prepare_folders(conf, *keys):
    """Prepare the folder mentioned in config under keys."""

    def makedir(folder):
        if not os.path.exists(folder):
            os.makedirs(folder)

    for key in keys:
        makedir(conf[key])


def load_global_config():
    """Load the global Software Heritage config"""

    return priority_read(
        swh_config_paths(SWH_GLOBAL_CONFIG),
        SWH_DEFAULT_GLOBAL_CONFIG,
    )


def load_named_config(name, default_conf=None, global_conf=True):
    """Load the config named `name` from the Software Heritage
    configuration paths.

    If global_conf is True (default), read the global configuration
    too.
    """

    conf = {}

    if global_conf:
        conf.update(load_global_config())

    conf.update(priority_read(swh_config_paths(name), default_conf))

    return conf


def load_from_envvar(default_config: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    """Load configuration yaml file from the environment variable SWH_CONFIG_FILENAME,
    eventually enriched with default configuration key/value from the default_config
    dict if provided.

    Returns:
        Configuration dict

    Raises:
        AssertionError if SWH_CONFIG_FILENAME is undefined

    """
    assert (
        "SWH_CONFIG_FILENAME" in os.environ
    ), "SWH_CONFIG_FILENAME environment variable is undefined."

    cfg_path = os.environ["SWH_CONFIG_FILENAME"]
    cfg = read_raw_config(cfg_path)
    cfg = merge_configs(default_config or {}, cfg)
    return cfg


def get_swh_backend_module(swh_package: str, cls: str) -> Tuple[str, Optional[type]]:
    entry_points = get_entry_points(group=f"swh.{swh_package}.classes")
    if not entry_points:
        # it's an "old-style" swh package, not declaring its classes entry point
        logger.warning(
            f"swh package does not yet declare the swh.{swh_package}.classes "
            "endpoint. Make sure all your swh dependencies are up to date."
        )
        if not swh_package.startswith("swh."):
            swh_package = f"swh.{swh_package}"
        return swh_package, None
    try:
        entry_point = entry_points[cls]
    except KeyError:
        raise ValueError(
            "Unknown %s class `%s`. Supported: %s"
            % (
                swh_package,
                cls,
                ", ".join(entry_point.name for entry_point in entry_points),
            )
        ) from None
    BackendCls = entry_point.load()
    return entry_point.module, BackendCls


def get_swh_backend_from_fullmodule(
    fullmodule: str,
) -> Tuple[Optional[str], Optional[str]]:
    if not fullmodule.startswith("swh."):
        fullmodule = f"swh.{fullmodule}"
    package = fullmodule.split(".")[1]

    entry_points = get_entry_points(group=f"swh.{package}.classes")
    for entry_point in entry_points:
        if entry_point.module == fullmodule:
            return package, entry_point.name

    return None, None


def list_swh_backends(package: str) -> List[str]:
    if package.startswith("swh."):
        package = package[4:]
    entry_points = get_entry_points(group=f"swh.{package}.classes")
    return [ep.name for ep in entry_points]


def list_db_config_entries(cfg) -> Generator[Tuple[str, str, dict, str], None, None]:
    """List all the db config entries in the given config structure

    Generates quadruplets (module, path, cfg, cnxstr) where:

    - the swh module name (aka top level config entries, eg. 'storage',
      'scheduler', etc.)

    - path: the path within the config structure of the (sub)config entry in
      which the db connection has been found,

    - cfg: the config subentry from the given gcfg in which the db config has
      been found; it contains at least a 'cls' key,

    - db: the db connection string

    """

    def look(cfg, path):
        if "cls" in cfg:
            for key, value in cfg.items():
                if key == "db" or key.endswith("_db"):
                    yield path, cfg, value
                elif isinstance(value, list):
                    for i, subcfg in enumerate(value):
                        yield from look(subcfg, path=f"{path}.{key}.{i}")
                elif isinstance(value, dict):
                    yield from look(value, path=f"{path}.{key}")

    for rootmodule, subcfg in cfg.items():
        for path, cfg_entry, cnxstr in look(subcfg, rootmodule):
            yield rootmodule, path, cfg_entry, cnxstr



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from importlib.metadata import distribution
import os

import swh.core.api.gunicorn_config as gunicorn_config


def test_post_fork_default(mocker):
    flask_integration = object()  # unique object to check for equality
    logging_integration = object()  # unique object to check for equality
    mocker.patch(
        "sentry_sdk.integrations.flask.FlaskIntegration", new=lambda: flask_integration
    )
    mocker.patch(
        "sentry_sdk.integrations.logging.LoggingIntegration",
        new=lambda event_level: logging_integration,
    )
    sentry_sdk_init = mocker.patch("sentry_sdk.init")

    gunicorn_config.post_fork(None, None)

    sentry_sdk_init.assert_called_once_with(
        dsn=None,
        integrations=[flask_integration, logging_integration],
        debug=False,
        release=None,
        environment=None,
    )


def test_post_fork_with_dsn_env(mocker):
    flask_integration = object()  # unique object to check for equality
    logging_integration = object()  # unique object to check for equality
    mocker.patch(
        "sentry_sdk.integrations.flask.FlaskIntegration", new=lambda: flask_integration
    )
    mocker.patch(
        "sentry_sdk.integrations.logging.LoggingIntegration",
        new=lambda event_level: logging_integration,
    )
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    mocker.patch.dict(os.environ, {"SWH_SENTRY_DSN": "test_dsn"})

    gunicorn_config.post_fork(None, None)

    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        integrations=[flask_integration, logging_integration],
        debug=False,
        release=None,
        environment=None,
    )


def test_post_fork_with_package_env(mocker):
    flask_integration = object()
    logging_integration = object()

    mocker.patch(
        "sentry_sdk.integrations.flask.FlaskIntegration", new=lambda: flask_integration
    )
    mocker.patch(
        "sentry_sdk.integrations.logging.LoggingIntegration",
        new=lambda event_level: logging_integration,
    )
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    mocker.patch.dict(
        os.environ,
        {
            "SWH_SENTRY_DSN": "test_dsn",
            "SWH_SENTRY_ENVIRONMENT": "tests",
            "SWH_MAIN_PACKAGE": "swh.core",
        },
    )

    gunicorn_config.post_fork(None, None)

    version = distribution("swh.core").version

    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        integrations=[flask_integration, logging_integration],
        debug=False,
        release="swh.core@" + version,
        environment="tests",
    )


def test_post_fork_debug(mocker):
    flask_integration = object()
    logging_integration = object()

    mocker.patch(
        "sentry_sdk.integrations.flask.FlaskIntegration", new=lambda: flask_integration
    )
    mocker.patch(
        "sentry_sdk.integrations.logging.LoggingIntegration",
        new=lambda event_level: logging_integration,
    )
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    mocker.patch.dict(
        os.environ, {"SWH_SENTRY_DSN": "test_dsn", "SWH_SENTRY_DEBUG": "1"}
    )

    gunicorn_config.post_fork(None, None)

    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        integrations=[flask_integration, logging_integration],
        debug=True,
        release=None,
        environment=None,
    )


def test_post_fork_no_flask(mocker):
    logging_integration = object()
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    mocker.patch(
        "sentry_sdk.integrations.logging.LoggingIntegration",
        new=lambda event_level: logging_integration,
    )
    mocker.patch.dict(os.environ, {"SWH_SENTRY_DSN": "test_dsn"})

    gunicorn_config.post_fork(None, None, flask=False)

    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        integrations=[logging_integration],
        debug=False,
        release=None,
        environment=None,
    )


def test_post_fork_override_logging_events_envvar(mocker):
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    mocker.patch.dict(
        os.environ,
        {"SWH_SENTRY_DSN": "test_dsn", "SWH_SENTRY_DISABLE_LOGGING_EVENTS": "false"},
    )

    gunicorn_config.post_fork(None, None, flask=False)

    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        integrations=[],
        debug=False,
        release=None,
        environment=None,
    )


def test_post_fork_extras(mocker):
    flask_integration = object()  # unique object to check for equality
    mocker.patch(
        "sentry_sdk.integrations.flask.FlaskIntegration", new=lambda: flask_integration
    )
    sentry_sdk_init = mocker.patch("sentry_sdk.init")
    mocker.patch.dict(os.environ, {"SWH_SENTRY_DSN": "test_dsn"})

    gunicorn_config.post_fork(
        None,
        None,
        sentry_integrations=["foo"],
        extra_sentry_kwargs={"bar": "baz"},
        disable_logging_events=False,
    )

    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        integrations=["foo", flask_integration],
        debug=False,
        bar="baz",
        release=None,
        environment=None,
    )



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from click.testing import CliRunner
from hypothesis import HealthCheck
import psycopg2
import pytest
from pytest_postgresql import factories

os.environ["LC_ALL"] = "C.UTF-8"

# we use getattr here to keep mypy happy regardless hypothesis version
function_scoped_fixture_check = (
    [getattr(HealthCheck, "function_scoped_fixture")]
    if hasattr(HealthCheck, "function_scoped_fixture")
    else []
)


def create_role_guest(**kwargs):
    with psycopg2.connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute("REVOKE CREATE ON SCHEMA public FROM PUBLIC")
            cur.execute("CREATE ROLE guest NOINHERIT LOGIN PASSWORD 'guest'")


postgresql_proc = factories.postgresql_proc(
    load=[create_role_guest],
)


@pytest.fixture
def cli_runner():
    return CliRunner()



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64
import datetime
from enum import Enum
import json
import traceback
import types
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from uuid import UUID

import iso8601
import msgpack
from requests import Response

from swh.core.api.classes import PagedResult


def encode_datetime(dt: datetime.datetime) -> str:
    """Wrapper of datetime.datetime.isoformat() that forbids naive datetimes."""
    if dt.tzinfo is None:
        raise TypeError("can not serialize naive 'datetime.datetime' object")
    return dt.isoformat()


def _encode_paged_result(obj: PagedResult) -> Dict[str, Any]:
    """Serialize PagedResult to a Dict."""
    return {
        "results": obj.results,
        "next_page_token": obj.next_page_token,
    }


def _decode_paged_result(obj: Dict[str, Any]) -> PagedResult:
    """Deserialize Dict into PagedResult"""
    return PagedResult(
        results=obj["results"],
        next_page_token=obj["next_page_token"],
    )


def exception_to_dict(exception: BaseException) -> Dict[str, Any]:
    tb = traceback.format_exception(None, exception, exception.__traceback__)
    exc_type = type(exception)
    return {
        "type": exc_type.__name__,
        "module": exc_type.__module__,
        "args": exception.args,
        "message": str(exception),
        "traceback": tb,
    }


def dict_to_exception(exc_dict: Dict[str, Any]) -> Exception:
    temp = __import__(exc_dict["module"], fromlist=[exc_dict["type"]])
    try:
        return getattr(temp, exc_dict["type"])(*exc_dict["args"])
    except Exception:
        # custom Exception type cannot be rebuilt, fallback to base Exception type
        return Exception(exc_dict["message"])


def encode_timedelta(td: datetime.timedelta) -> Dict[str, int]:
    return {
        "days": td.days,
        "seconds": td.seconds,
        "microseconds": td.microseconds,
    }


ENCODERS: List[Tuple[type, str, Callable]] = [
    (UUID, "uuid", str),
    (datetime.timedelta, "timedelta", encode_timedelta),
    (PagedResult, "paged_result", _encode_paged_result),
    (Exception, "exception", exception_to_dict),
]

JSON_ENCODERS: List[Tuple[type, str, Callable]] = [
    (datetime.datetime, "datetime", encode_datetime),
    (bytes, "bytes", lambda o: base64.b85encode(o).decode("ascii")),
]

DECODERS: Dict[str, Callable] = {
    "timedelta": lambda d: datetime.timedelta(**d),
    "uuid": UUID,
    "paged_result": _decode_paged_result,
    "exception": dict_to_exception,
    # for BW compat, to be moved in JSON_DECODERS ASAP
    "datetime": lambda d: iso8601.parse_date(d, default_timezone=None),
}

JSON_DECODERS: Dict[str, Callable] = {
    "bytes": base64.b85decode,
}


def get_encoders(
    extra_encoders: Optional[List[Tuple[Type, str, Callable]]], with_json: bool = False
) -> List[Tuple[Type, str, Callable]]:
    encoders = ENCODERS
    if with_json:
        encoders = [*encoders, *JSON_ENCODERS]
    if extra_encoders:
        encoders = [*encoders, *extra_encoders]
    return encoders


def get_decoders(
    extra_decoders: Optional[Dict[str, Callable]], with_json: bool = False
) -> Dict[str, Callable]:
    decoders = DECODERS
    if with_json:
        decoders = {**decoders, **JSON_DECODERS}
    if extra_decoders is not None:
        decoders = {**decoders, **extra_decoders}
    return decoders


class MsgpackExtTypeCodes(Enum):
    LONG_INT = 1
    LONG_NEG_INT = 2


def encode_data_client(data: Any, extra_encoders=None) -> bytes:
    try:
        return msgpack_dumps(data, extra_encoders=extra_encoders)
    except OverflowError as e:
        raise ValueError("Limits were reached. Please, check your input.\n" + str(e))


def decode_response(response: Response, extra_decoders=None) -> Any:
    content_type = response.headers["content-type"]

    if content_type.startswith("application/x-msgpack"):
        r = msgpack_loads(response.content, extra_decoders=extra_decoders)
    elif content_type.startswith("application/json"):
        r = json_loads(response.text, extra_decoders=extra_decoders)
    elif content_type.startswith("text/"):
        r = response.text
    else:
        raise ValueError("Wrong content type `%s` for API response" % content_type)

    return r


class SWHJSONEncoder(json.JSONEncoder):
    """JSON encoder for data structures generated by Software Heritage.

    This JSON encoder extends the default Python JSON encoder and adds
    awareness for the following specific types:

    - bytes (get encoded as a Base85 string);
    - datetime.datetime (get encoded as an ISO8601 string).

    Non-standard types get encoded as a a dictionary with two keys:

    - swhtype with value 'bytes' or 'datetime';
    - d containing the encoded value.

    SWHJSONEncoder also encodes arbitrary iterables as a list
    (allowing serialization of generators).

    Caveats: Limitations in the JSONEncoder extension mechanism
    prevent us from "escaping" dictionaries that only contain the
    swhtype and d keys, and therefore arbitrary data structures can't
    be round-tripped through SWHJSONEncoder and SWHJSONDecoder.

    """

    def __init__(self, extra_encoders=None, **kwargs):
        super().__init__(**kwargs)
        self.encoders = get_encoders(extra_encoders, with_json=True)

    def default(self, o: Any) -> Union[Dict[str, Union[Dict[str, int], str]], list]:
        for type_, type_name, encoder in self.encoders:
            if isinstance(o, type_):
                return {
                    "swhtype": type_name,
                    "d": encoder(o),
                }
        try:
            return super().default(o)
        except TypeError as e:
            try:
                iterable = iter(o)
            except TypeError:
                raise e from None
            else:
                return list(iterable)


class SWHJSONDecoder(json.JSONDecoder):
    """JSON decoder for data structures encoded with SWHJSONEncoder.

    This JSON decoder extends the default Python JSON decoder,
    allowing the decoding of:

    - bytes (encoded as a Base85 string);
    - datetime.datetime (encoded as an ISO8601 string).

    Non-standard types must be encoded as a a dictionary with exactly
    two keys:

    - swhtype with value 'bytes' or 'datetime';
    - d containing the encoded value.

    To limit the impact our encoding, if the swhtype key doesn't
    contain a known value, the dictionary is decoded as-is.

    """

    def __init__(self, extra_decoders=None, **kwargs):
        super().__init__(**kwargs)
        self.decoders = get_decoders(extra_decoders, with_json=True)

    def decode_data(self, o: Any) -> Any:
        if isinstance(o, dict):
            if set(o.keys()) == {"d", "swhtype"}:
                if o["swhtype"] == "bytes":
                    return base64.b85decode(o["d"])
                decoder = self.decoders.get(o["swhtype"])
                if decoder:
                    return decoder(self.decode_data(o["d"]))
            return {key: self.decode_data(value) for key, value in o.items()}
        if isinstance(o, list):
            return [self.decode_data(value) for value in o]
        else:
            return o

    def raw_decode(self, s: str, idx: int = 0) -> Tuple[Any, int]:
        data, index = super().raw_decode(s, idx)
        return self.decode_data(data), index


def json_dumps(data: Any, extra_encoders=None) -> str:
    return json.dumps(data, cls=SWHJSONEncoder, extra_encoders=extra_encoders)


def json_loads(data: str, extra_decoders=None) -> Any:
    return json.loads(data, cls=SWHJSONDecoder, extra_decoders=extra_decoders)


def msgpack_dumps(data: Any, extra_encoders=None) -> bytes:
    """Write data as a msgpack stream"""
    encoders = get_encoders(extra_encoders)

    def encode_types(obj):
        if isinstance(obj, int):
            # integer overflowed while packing. Handle it as an extended type
            if obj > 0:
                code = MsgpackExtTypeCodes.LONG_INT.value
            else:
                code = MsgpackExtTypeCodes.LONG_NEG_INT.value
                obj = -obj
            length, rem = divmod(obj.bit_length(), 8)
            if rem:
                length += 1
            return msgpack.ExtType(code, int.to_bytes(obj, length, "big"))

        if isinstance(obj, types.GeneratorType):
            return list(obj)

        for type_, type_name, encoder in encoders:
            if isinstance(obj, type_):
                return {
                    b"swhtype": type_name,
                    b"d": encoder(obj),
                }
        return obj

    return msgpack.packb(
        data,
        use_bin_type=True,
        datetime=True,  # encode datetime as msgpack.Timestamp
        default=encode_types,
    )


def msgpack_loads(data: bytes, extra_decoders=None) -> Any:
    """Read data as a msgpack stream.

    .. Caution::
       This function is used by swh.journal to decode the contents of the
       journal. This function **must** be kept backwards-compatible.
    """
    decoders = get_decoders(extra_decoders)

    def ext_hook(code, data):
        if code == MsgpackExtTypeCodes.LONG_INT.value:
            return int.from_bytes(data, "big")
        elif code == MsgpackExtTypeCodes.LONG_NEG_INT.value:
            return -int.from_bytes(data, "big")
        raise ValueError("Unknown msgpack extended code %s" % code)

    def decode_types(obj):
        # Support for current encodings
        if set(obj.keys()) == {b"d", b"swhtype"}:
            decoder = decoders.get(obj[b"swhtype"])
            if decoder:
                return decoder(obj[b"d"])

        # Fallthrough
        return obj

    return msgpack.unpackb(
        data,
        raw=False,
        object_hook=decode_types,
        ext_hook=ext_hook,
        strict_map_key=False,
        timestamp=3,  # convert Timestamp in datetime objects (tz UTC)
    )



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from click.testing import CliRunner
import pytest

from swh.core.tests.test_cli import assert_result


@pytest.fixture
def swhmain(swhmain):
    from swh.core.cli.backend import backend as swhbackend

    swhmain.add_command(swhbackend)
    return swhmain


def test_backend_list_ok(swhmain, mock_get_entry_points):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["backend", "list", "test"])
    assert_result(result)
    assert result.output.strip() == "backend1 A mockup backend for tests"


def test_backend_list_empty(swhmain, mock_get_entry_points):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["backend", "list", "wrong_package"])
    assert result.exit_code == 1
    assert "No backend found for package 'wrong_package'" in result.output.strip()


def test_backend_list_cls_ok(swhmain, mock_get_entry_points):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["backend", "list", "test", "backend1"])
    assert_result(result)
    assert result.output.strip() == "test:backend1\n\nA mockup backend for tests"


def test_backend_list_cls_no_package(swhmain, mock_get_entry_points):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["backend", "list", "wrong_package", "backend1"])
    assert result.exit_code == 1
    assert (
        "No backend 'backend1' found for package 'wrong_package'"
        in result.output.strip()
    )


def test_backend_list_cls_no_backend(swhmain, mock_get_entry_points):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["backend", "list", "test", "backend2"])
    assert result.exit_code == 1
    assert "No backend 'backend2' found for package 'test'" in result.output.strip()



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from subprocess import CalledProcessError

import pytest
from pytest_postgresql import factories
import yaml

from swh.core.cli.db import db as swhdb
from swh.core.db import BaseDb
from swh.core.db.db_utils import swh_db_module, swh_db_version
from swh.core.tests.test_cli import assert_result, assert_section_contains

postgresql2 = factories.postgresql("postgresql_proc", dbname="tests2")


def test_cli_swh_help(swhmain, cli_runner):
    swhmain.add_command(swhdb)
    result = cli_runner.invoke(swhmain, ["-h"])
    assert_result(result)
    assert_section_contains(
        result.output, "Commands", "db  Software Heritage database generic tools."
    )


help_db_snippets = (
    (
        "Usage",
        (
            "Usage: swh db [OPTIONS] COMMAND [ARGS]...",
            "Software Heritage database generic tools.",
        ),
    ),
    (
        "Commands",
        (
            "create      Create a database for the Software Heritage <module>.",
            "init        Initialize a database for the Software Heritage <module>.",
            "init-admin  Execute superuser-level initialization steps",
        ),
    ),
)


def test_cli_swh_db_help(swhmain, cli_runner):
    swhmain.add_command(swhdb)
    result = cli_runner.invoke(swhmain, ["db", "-h"])
    assert_result(result)
    for section, snippets in help_db_snippets:
        for snippet in snippets:
            assert_section_contains(result.output, section, snippet)


@pytest.fixture
def swh_db_cli(cli_runner, monkeypatch, postgresql):
    """This initializes a cli_runner and sets the correct environment variable expected by
    the cli to run appropriately (when not specifying the --dbname flag)

    """
    monkeypatch.setenv("PGHOST", postgresql.info.host)
    monkeypatch.setenv("PGUSER", postgresql.info.user)
    monkeypatch.setenv("PGPORT", str(postgresql.info.port))

    return cli_runner, postgresql.info


def craft_conninfo(test_db, dbname=None) -> str:
    """Craft conninfo string out of the test_db object. This also allows to override the
    dbname."""
    db_params = test_db.info
    dbname = dbname if dbname else db_params.dbname
    return f"postgresql://{db_params.user}@{db_params.host}:{db_params.port}/{dbname}"


@pytest.mark.parametrize(
    "module_table",
    [
        ("test", "origin"),
        ("test.postgresql", "origin"),
        ("test:postgresql", "origin"),
        ("test:cli2", "origin2"),
        ("test.cli2", "origin2"),
    ],
)
def test_cli_swh_db_create_and_init_db(
    cli_runner, postgresql, mock_get_entry_points, module_table
):
    """Create a db then initializing it should be ok"""
    module_name, table = module_table
    conninfo = craft_conninfo(postgresql, f"db-{module_name}")
    # This creates the db and installs the necessary admin extensions
    result = cli_runner.invoke(swhdb, ["create", module_name, "--dbname", conninfo])
    assert_result(result)

    # This initializes the schema and data
    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    assert_result(result)

    # the origin value in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, installed during db creation step)
    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute(f"select * from {table}")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_initialization_fail_without_creation_first(
    cli_runner, postgresql, mock_import_module
):
    """Init command on an inexisting db cannot work"""
    module_name = "test"  # it's mocked here
    conninfo = craft_conninfo(postgresql, "inexisting-db")

    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    # Fails because we cannot connect to an inexisting db
    assert result.exit_code == 1, f"Unexpected output: {result.output}"
    assert isinstance(result.exception, CalledProcessError)
    assert b'FATAL:  database "inexisting-db" does not exist' in result.exception.stderr


def test_cli_swh_db_initialization_fail_without_extension(
    cli_runner, postgresql, mock_import_module
):
    """Init command cannot work without privileged extension.

    In this test, the schema needs privileged extension to work.

    """
    module_name = "test.postgresql"  # it's mocked here
    conninfo = craft_conninfo(postgresql)

    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    # Fails as the function `public.digest` is not installed, init-admin calls is needed
    # first (the next tests show such behavior)
    assert result.exit_code == 1, f"Unexpected output: {result.output}"
    assert isinstance(result.exception, CalledProcessError)
    assert (
        b"ERROR:  function public.digest(text, unknown) does not exist"
        in result.exception.stderr
    )


def test_cli_swh_db_initialization_works_with_flags(
    cli_runner,
    postgresql,
    mock_get_entry_points,
):
    """Init commands with carefully crafted libpq conninfo works"""
    module_name = "test"  # it's mocked here
    conninfo = craft_conninfo(postgresql)

    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert_result(result)

    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    assert_result(result)
    # the origin values in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, init-admin calls installs it)
    with BaseDb.connect(postgresql.info.dsn).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_initialization_with_env(
    swh_db_cli, mock_get_entry_points, postgresql
):
    """Init commands with standard environment variables works"""
    module_name = "test"  # it's mocked here
    cli_runner, db_params = swh_db_cli
    result = cli_runner.invoke(
        swhdb, ["init-admin", module_name, "--dbname", db_params.dbname]
    )
    assert_result(result)
    result = cli_runner.invoke(
        swhdb,
        [
            "init",
            module_name,
            "--dbname",
            db_params.dbname,
        ],
    )
    assert_result(result)

    # the origin values in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, init-admin calls installs it)
    with BaseDb.connect(postgresql.info.dsn).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_initialization_idempotent(
    swh_db_cli, mock_get_entry_points, postgresql
):
    """Multiple runs of the init commands are idempotent"""
    module_name = "test"  # mocked
    cli_runner, db_params = swh_db_cli

    result = cli_runner.invoke(
        swhdb, ["init-admin", module_name, "--dbname", db_params.dbname]
    )
    assert_result(result)

    result = cli_runner.invoke(
        swhdb, ["init", module_name, "--dbname", db_params.dbname]
    )
    assert_result(result)

    result = cli_runner.invoke(
        swhdb, ["init-admin", module_name, "--dbname", db_params.dbname]
    )
    assert_result(result)

    result = cli_runner.invoke(
        swhdb, ["init", module_name, "--dbname", db_params.dbname]
    )
    assert_result(result)

    # the origin values in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, init-admin calls installs it)
    with BaseDb.connect(postgresql.info.dsn).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


@pytest.mark.parametrize("with_module_config_key", [True, False])
def test_cli_swh_db_create_and_init_db_new_api(
    cli_runner,
    postgresql,
    mock_get_entry_points,
    mocker,
    tmp_path,
    with_module_config_key,
):
    """Create a db then initializing it should be ok for a "new style" datastore"""
    module_name = "test"

    conninfo = craft_conninfo(postgresql)

    # This initializes the schema and data
    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(yaml.dump({module_name: {"cls": "postgresql", "db": conninfo}}))
    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert_result(result)

    cli_cmd = ["-C", cfgfile, "init", module_name]
    if with_module_config_key:
        cli_cmd.extend(["--module-config-key", module_name])
    result = cli_runner.invoke(swhdb, cli_cmd)
    assert_result(result)

    # the origin value in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, installed during db creation step)
    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


@pytest.mark.init_version(version=2)
def test_cli_swh_db_upgrade_new_api(
    request,
    cli_runner,
    mock_get_entry_points,
    postgresql,
    datadir,
    mocker,
    tmp_path,
):
    """Upgrade scenario for a "new style" datastore"""
    module_name = "test"

    current_version = request.node.get_closest_marker("init_version").kwargs["version"]

    conninfo = craft_conninfo(postgresql)

    # This initializes the schema and data
    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(yaml.dump({module_name: {"cls": "postgresql", "db": conninfo}}))
    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", module_name])
    assert_result(result)

    assert swh_db_version(conninfo) == 2

    # the upgrade should not do anything because the datastore does advertise
    # version 1
    current_version = 1
    request.node.get_closest_marker("init_version").kwargs["version"] = current_version
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", module_name])
    assert_result(result)
    assert swh_db_version(conninfo) == 2

    # advertise current version as 3, a simple upgrade should get us there, but
    # no further
    current_version = 3
    request.node.get_closest_marker("init_version").kwargs["version"] = current_version
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", module_name])
    assert_result(result)
    assert swh_db_version(conninfo) == 3

    # an attempt to go further should generate an error
    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "upgrade", module_name, "--to-version", 5]
    )
    assert result.exit_code != 0
    assert swh_db_version(conninfo) == 3
    # an attempt to go lower should not do anything
    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "upgrade", module_name, "--to-version", 2]
    )
    assert_result(result)
    assert swh_db_version(conninfo) == 3

    # advertise current version as 6, an upgrade with --to-version 4 should
    # stick to the given version 4 and no further
    current_version = 6
    request.node.get_closest_marker("init_version").kwargs["version"] = current_version
    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "upgrade", module_name, "--to-version", 4]
    )
    assert_result(result)
    assert swh_db_version(conninfo) == 4
    assert "migration was not complete" in result.output

    # attempt to upgrade to a newer version than current code version fails
    result = cli_runner.invoke(
        swhdb,
        ["-C", cfgfile, "upgrade", module_name, "--to-version", current_version + 1],
    )
    assert result.exit_code != 0
    assert swh_db_version(conninfo) == 4

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.execute("drop table dbmodule")
    assert swh_db_module(conninfo) is None

    # db migration should recreate the missing dbmodule table
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", module_name])
    assert_result(result)
    assert "Warning: the database does not have a dbmodule table." in result.output
    assert (
        "Write the module information (test:postgresql) in the database? [Y/n]"
        in result.output
    )
    assert swh_db_module(conninfo) == "test:postgresql"


@pytest.mark.init_version(version=5)
def test_cli_swh_db_init_version_ok(
    request,
    cli_runner,
    mock_get_entry_points,
    postgresql,
    datadir,
    mocker,
    tmp_path,
):
    """Upgrade scenario for a "new style" datastore"""
    module_name = "test"

    # the `current_version` variable is the version that will be returned by
    # any call to `get_current_version()` in this test session, thanks to the
    # local mocked version of import_swhmodule() below.
    current_version = request.node.get_closest_marker("init_version").kwargs["version"]
    conninfo = craft_conninfo(postgresql)

    # call the db init stuff WITHOUT a config file
    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    assert_result(result)

    assert swh_db_version(conninfo) == current_version


def test_cli_swh_db_version(swh_db_cli, mock_get_entry_points, postgresql):
    module_name = "test"
    cli_runner, db_params = swh_db_cli

    conninfo = craft_conninfo(postgresql, "test-db-version")
    # This creates the db and installs the necessary admin extensions
    result = cli_runner.invoke(swhdb, ["create", module_name, "--dbname", conninfo])
    assert_result(result)
    # This initializes the schema and data
    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    assert_result(result)

    actual_db_version = swh_db_version(conninfo)

    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute("select version from dbversion order by version desc limit 1")
        expected_version = cur.fetchone()[0]
        assert actual_db_version == expected_version

    assert_result(result)
    assert (
        f"initialized (flavor default) at version {expected_version}" in result.output
    )


@pytest.mark.parametrize("initialize_all", [True, False])
def test_cli_swh_db_initadmin_and_init_db_from_config_path(
    cli_runner,
    postgresql,
    postgresql2,
    mock_get_entry_points,
    mocker,
    tmp_path,
    initialize_all,
):
    """Test init-admin and init commands with db cnx string coming from the config file

    It will test both the case where db cnx location in the config file are
    given and the automated mode (aka with --initialize-all).

    """
    conninfo = craft_conninfo(postgresql)
    conninfo2 = craft_conninfo(postgresql2)

    # This initializes the schema and data
    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(
        f"""
test:
  cls: something
  backend:
    cls: pipeline
    steps:
      - cls: postgresql
        db: {conninfo}
      - cls: stuff
        backend:
          cls: cli2
          cli_db: {conninfo2}
    """
    )
    if initialize_all:
        result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init-admin", "-a", "test"])
        assert_result(result)
        result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", "-a", "test"])
        assert_result(result)
    else:
        for config_path in (
            "test.backend.steps.0",
            "test.backend.steps.1.backend",
        ):
            result = cli_runner.invoke(
                swhdb, ["-C", cfgfile, "init-admin", "-p", config_path]
            )
            assert_result(result)
            result = cli_runner.invoke(
                swhdb, ["-C", cfgfile, "init", "-p", config_path]
            )
            assert_result(result)

    # the origin value in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, installed during db creation step)
    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1

    # same with the origin2 table
    with BaseDb.connect(conninfo2).cursor() as cur:
        cur.execute("select * from origin2")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_list_config_path(
    cli_runner,
    postgresql,
    postgresql2,
    mock_get_entry_points,
    mocker,
    tmp_path,
):
    """Test the swh db list command"""
    conninfo = craft_conninfo(postgresql)
    conninfo2 = craft_conninfo(postgresql2)

    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(
        f"""
test:
  cls: something
  backend:
    cls: pipeline
    steps:
      - cls: postgresql
        db: {conninfo}
      - cls: cli
        backend:
          cls: cli2
          cli_db: {conninfo2}
    """
    )
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "list"])
    assert_result(result)
    assert (
        result.output
        == f"""\
test.backend.steps.0 postgresql {conninfo}
test.backend.steps.1.backend cli2 {conninfo2}
"""
    )


def test_cli_swh_db_version_from_config(
    cli_runner,
    postgresql,
    postgresql2,
    mock_get_entry_points,
    mocker,
    tmp_path,
):
    """Test the swh db list command"""
    conninfo = craft_conninfo(postgresql)
    conninfo2 = craft_conninfo(postgresql2)

    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(
        f"""
test:
  cls: something
  backend:
    cls: pipeline
    steps:
      - cls: postgresql
        db: {conninfo}
      - cls: cli
        backend:
          cls: cli2
          db: {conninfo2}
    """
    )
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init-admin", "-a", "test"])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", "-a", "test"])
    assert_result(result)

    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "version", "test"])
    # this one should fail, there is no "natural" config entry in this config
    # file for the 'test' module
    assert result.exit_code != 0

    # but we can ask for each entry
    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "version", "-p", "test.backend.steps.0"]
    )
    assert_result(result)
    assert (
        result.output
        == """
module: test:postgresql
flavor: default
current code version: 3
version: 3
"""
    )

    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "version", "-p", "test.backend.steps.1.backend"]
    )
    assert_result(result)
    assert (
        result.output
        == """
module: test:cli2
current code version: 3
version: 3
"""
    )

    # or all at once
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "version", "--all", "test"])
    assert_result(result)
    assert (
        result.output
        == """
module: test:postgresql
flavor: default
current code version: 3
version: 3

module: test:cli2
current code version: 3
version: 3
"""
    )


@pytest.mark.init_version(version=1)
def test_cli_swh_db_upgrade_from_config(
    request,
    cli_runner,
    mock_get_entry_points,
    postgresql,
    postgresql2,
    datadir,
    mocker,
    tmp_path,
):
    """Test the upgrade cli tool reading db cnx from a nested config file"""
    conninfo = craft_conninfo(postgresql)
    conninfo2 = craft_conninfo(postgresql2)

    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(
        f"""
test:
  cls: something
  backend:
    cls: pipeline
    steps:
      - cls: postgresql
        db: {conninfo}
      - cls: cli
        backend:
          cls: cli2
          cli_db: {conninfo2}
    """
    )

    module_name = "test"

    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init-admin", "-a", module_name])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", "-a", module_name])
    assert_result(result)

    assert swh_db_version(conninfo) == 1
    assert swh_db_version(conninfo2) == 1

    for module_name, config_path, cnxstr in (
        ("test:postgresql", "test.backend.steps.0", conninfo),
        ("test:cli2", "test.backend.steps.1.backend", conninfo2),
    ):
        current_version = 1
        # XXX hack hack hack: change the current test (pytest.)marker's
        # init_version arg, this one is used in mock_import_swhmodule...
        request.node.get_closest_marker("init_version").kwargs[
            "version"
        ] = current_version
        # the upgrade should not do anything because the datastore does advertise
        # version 1
        result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", "-p", config_path])
        assert_result(result)
        assert swh_db_version(cnxstr) == 1

        # advertise current version as 3, a simple upgrade should get us there, but
        # no further
        current_version = 2
        request.node.get_closest_marker("init_version").kwargs[
            "version"
        ] = current_version
        result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", "-p", config_path])
        assert_result(result)
        assert swh_db_version(cnxstr) == 2

        # an attempt to go further should not do anything
        result = cli_runner.invoke(
            swhdb, ["-C", cfgfile, "upgrade", "-p", config_path, "--to-version", 5]
        )
        assert result.exit_code != 0
        assert swh_db_version(cnxstr) == 2
        # an attempt to go lower should not do anything
        result = cli_runner.invoke(
            swhdb, ["-C", cfgfile, "upgrade", "-p", config_path, "--to-version", 1]
        )
        assert_result(result)
        assert swh_db_version(cnxstr) == 2

        # advertise current version as 6, an upgrade with --to-version 4 should
        # stick to the given version 4 and no further
        current_version = 6
        request.node.get_closest_marker("init_version").kwargs[
            "version"
        ] = current_version
        result = cli_runner.invoke(
            swhdb, ["-C", cfgfile, "upgrade", "-p", config_path, "--to-version", 4]
        )
        assert_result(result)
        assert swh_db_version(cnxstr) == 4
        assert "migration was not complete" in result.output

        # attempt to upgrade to a newer version than current code version fails
        result = cli_runner.invoke(
            swhdb,
            [
                "-C",
                cfgfile,
                "upgrade",
                config_path,
                "--to-version",
                current_version + 1,
            ],
        )
        assert result.exit_code != 0
        assert swh_db_version(cnxstr) == 4

        cnx = BaseDb.connect(cnxstr)
        with cnx.transaction() as cur:
            cur.execute("drop table dbmodule")
        assert swh_db_module(cnxstr) is None

        # db migration should recreate the missing dbmodule table
        result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", "-p", config_path])
        assert_result(result)
        assert "Warning: the database does not have a dbmodule table." in result.output
        assert (
            f"Write the module information ({module_name}) in the database? [Y/n]"
            in result.output
        )
        assert swh_db_module(cnxstr) == module_name


@pytest.mark.init_version(version=1)
def test_cli_swh_db_upgrade_all(
    request,
    cli_runner,
    mock_get_entry_points,
    postgresql,
    postgresql2,
    datadir,
    mocker,
    tmp_path,
):
    """Test the upgrade cli tool reading db cnx from a nested config file"""
    conninfo = craft_conninfo(postgresql)
    conninfo2 = craft_conninfo(postgresql2)

    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(
        f"""
test:
  cls: something
  backend:
    cls: pipeline
    steps:
      - cls: postgresql
        db: {conninfo}
      - cls: cli
        backend:
          cls: cli2
          db: {conninfo2}
    """
    )

    module_name = "test"

    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init-admin", "-a", module_name])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", "-a", module_name])
    assert_result(result)

    assert swh_db_version(conninfo) == 1
    assert swh_db_version(conninfo2) == 1

    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", "-a", module_name])
    assert_result(result)
    assert swh_db_version(conninfo) == 1
    assert swh_db_version(conninfo2) == 1

    current_version = 6
    request.node.get_closest_marker("init_version").kwargs["version"] = current_version

    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", "-a", module_name])
    assert_result(result)
    assert swh_db_version(conninfo) == 6
    assert swh_db_version(conninfo2) == 6



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os

import pytest
from yaml import safe_load

from swh.core.logging import BadLogLevel, logging_configure, validate_loglevel


def test_logging_configure_default():
    """Logging should be configured to INFO by default"""
    root_log_level = logging_configure()

    assert root_log_level == "INFO"


def test_logging_configure_with_override():
    """Logging module should be configured according to log_levels provided."""
    log_levels = [
        ("swh.core.tests", logging.DEBUG),
        ("swh.core.api", logging.CRITICAL),
        ("swh.core.db", logging.ERROR),
        ("swh.loader.core.tests", logging.ERROR),
        ("swh.loader.core", logging.WARNING),
    ]
    for module, expected_log_level in log_levels:
        logger = logging.getLogger(module)
        assert logger.getEffectiveLevel() != expected_log_level

    # Set it up
    root_log_level = logging_configure(log_levels)
    assert root_log_level == "INFO"

    for module, expected_log_level in log_levels:
        logger = logging.getLogger(module)
        assert logger.getEffectiveLevel() == expected_log_level


def test_logging_configure_from_yaml(datadir):
    """Logging should be configurable from yaml configuration file."""
    logging_config = os.path.join(datadir, "logging-config.yaml")
    root_log_level = logging_configure([], logging_config)

    with open(logging_config, "r") as f:
        config = safe_load(f.read())

    for module, logger_config in config["loggers"].items():
        if not logger_config:
            continue
        log_level = logger_config["level"]

        logger = logging.getLogger(module)
        assert logger.getEffectiveLevel() == logging.getLevelName(log_level)

    assert root_log_level == config["root"]["level"]


@pytest.mark.parametrize(
    "log_level,expected_module,expected_log_level",
    [
        ("swh.core:DEBUG", "swh.core", logging.DEBUG),
        ("swh.core:debug", "swh.core", logging.DEBUG),
        ("swh.core.api:info", "swh.core.api", logging.INFO),
    ],
)
def test_validate_loglevel_ok(log_level, expected_module, expected_log_level):
    """Correct log level should pass validation"""
    module, level = validate_loglevel(log_level)

    assert module == expected_module
    assert level == expected_log_level


def test_validate_loglevel_raise():
    """Unsupported log level should raise"""

    with pytest.raises(BadLogLevel, match="unknown"):
        validate_loglevel("inexistent")

    with pytest.raises(BadLogLevel, match="unknown"):
        validate_loglevel("swh.core:something-wrong")



#!/usr/bin/env python3
# Copyright (C) 2018-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from os import environ
from typing import Any, Dict, List, Optional, Tuple
import warnings

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

warnings.filterwarnings("ignore")  # noqa prevent psycopg from side-tracking us


logger = logging.getLogger(__name__)


@swh_cli_group.group(name="db", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Configuration file.",
)
@click.pass_context
def db(ctx, config_file):
    """Software Heritage database generic tools."""
    from swh.core.config import read as config_read

    ctx.ensure_object(dict)
    if config_file is None:
        config_file = environ.get("SWH_CONFIG_FILENAME")
    cfg = config_read(config_file)
    ctx.obj["config"] = cfg


@db.command(name="create", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name.",
    default=None,
)
@click.option(
    "--template",
    "-T",
    help="Template database from which to build this database.",
    default="template1",
    show_default=True,
)
@click.pass_context
def db_create(ctx, module, dbname, template):
    """Create a database for the Software Heritage <module>.

    and potentially execute superuser-level initialization steps.

    Example::

        swh db create -d swh-test storage

    If you want to specify non-default postgresql connection parameters, please
    provide them using standard environment variables or by the mean of a
    properly crafted libpq connection URI. See psql(1) man page (section
    ENVIRONMENTS) for details.

    Note: this command requires a postgresql connection with superuser permissions.

    Example::

        \b
        PGPORT=5434 swh db create indexer
        swh db create -d postgresql://superuser:passwd@pghost:5433/swh-storage storage

    """
    from swh.core.db.db_utils import create_database_for_package

    args = handle_cmd_args(
        cfg=ctx.obj["config"],
        module=module,
        do_all=False,
        dbname=dbname,
        is_path=False,
    )

    for package, fullmodule, backend_class, dbname, cfg in args:
        create_database_for_package(fullmodule, dbname, template)


@db.command(name="init-admin", context_settings=CONTEXT_SETTINGS)
@click.argument("module", metavar="MODULE-OR-CONFIG-PATH", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name.",
    default=None,
)
@click.option(
    "-a",
    "--all",
    "initialize_all",
    help="superuser initialize all db found in the config file for the swh 'module'",
    default=False,
    is_flag=True,
)
@click.option(
    "-p",
    "--module-is-path",
    "module_is_path",
    help="If set, inpterpret the given 'module' as a config 'json path' within the config file",
    default=False,
    is_flag=True,
)
@click.pass_context
def db_init_admin(
    ctx, module: str, dbname: Optional[str], initialize_all: bool, module_is_path: bool
) -> None:
    """Execute superuser-level initialization steps (e.g pg extensions, admin functions,
    ...)

    Note: this command requires a postgresql connection with superuser permissions (e.g
    postgres, swh-admin, ...)

    If given, a db connection string will be used to connect to the database
    and execute the initialization steps for the given module.

    Example::

        PGPASSWORD=... swh db init-admin -d swh-test scheduler

    If you want to specify non-default postgresql connection parameters, you can
    provide them using standard environment variables or by the mean of a
    properly crafted libpq connection URI. See psql(1) man page (section
    ENVIRONMENTS) for details.

    Examples::

        \b
        PGPORT=5434 swh db init-admin scheduler
        swh db init-admin -d postgresql://superuser:passwd@pghost:5433/swh-scheduler \
          scheduler

    If the db connection string is not given, it will be looked for in the
    configuration file. Note that this step needs admin right on the database,
    so this usage is not meant for production environment (but rather test
    environments.)

    Example::

        \b
        $ cat conf.yml
        storage:
          cls: postgresql
          db: postgresql://user:passwd@pghost:5433/swh-storage
          objstorage:
            cls: memory

        \b
        $ swh db -C conf.yml init-admin storage

    With '-p', the module is interpreted as a 'path' in the configuration file
    where the configuration entry for the targeted database connection string
    can be found. For example::

        \b
        $ cat conf.yml
        storage:
          cls: pipeline
          steps:
            - cls: masking
              masking_db: postgresql:///?service=swh-masking-proxy
            - cls: buffer
            - cls: postgresql
              db: postgresql://user:passwd@pghost:5433/swh-storage
              objstorage:
                cls: memory

        \b
        $ swh db -C conf.yml init-admin -p storage.steps.2

    The --all option allows to execute superuser-level
    initialization steps for all the datasabases found in the config file for
    the <module>. For example::

        \b
        $ cat conf.yml
        storage:
          cls: pipeline
          steps:
            - cls: masking
              masking_db: postgresql:///?service=swh-masking-proxy
            - cls: buffer
            - cls: postgresql
              db: postgresql://user:passwd@pghost:5433/swh-storage
              objstorage:
                cls: memory

        \b
        $ swh db -C conf.yml init-admin -a storage

    will run the superuser-level init for both the masking and main storage
    declared in the 'storage' section of the config file.

    """
    from swh.core.db.db_utils import init_admin_extensions

    args = handle_cmd_args(
        cfg=ctx.obj["config"],
        module=module,
        do_all=initialize_all,
        dbname=dbname,
        is_path=module_is_path,
    )

    for package, fullmodule, _, dbname, cfg in args:
        logger.debug("db_init_admin %s:%s dbname=%s", package, cfg["cls"], dbname)
        init_admin_extensions(f"{package}:{cfg['cls']}", dbname)


@db.command(name="list", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=False)
@click.pass_context
def db_list(ctx, module):
    """List found DB configs under the <module> in the config file"""
    from swh.core.config import list_db_config_entries

    cfg = ctx.obj["config"]
    for swhmod, path, dbcfg, db in list_db_config_entries(cfg):
        if module and module != swhmod:
            continue
        print(path, dbcfg["cls"], db)


@db.command(name="init", context_settings=CONTEXT_SETTINGS)
@click.argument("module", metavar="MODULE-OR-CONFIG-PATH", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name or connection URI.",
    default=None,
    show_default=False,
)
@click.option(
    "--flavor",
    help="Database flavor.",
    default=None,
)
@click.option("--module-config-key", help="Module config key to lookup.", default=None)
@click.option(
    "-a",
    "--all",
    "initialize_all",
    help="initialize all db found in the config file for the swh 'module'",
    default=False,
    is_flag=True,
)
@click.option(
    "-p",
    "--module-is-path",
    "module_is_path",
    help="If set, inpterpret the given 'module' as a config 'json path' within the config file",
    default=False,
    is_flag=True,
)
@click.pass_context
def db_init(
    ctx, module, dbname, flavor, module_config_key, initialize_all, module_is_path
):
    """Initialize a database for the Software Heritage <module>.

    As for the 'init-admin' command, the database connection string can come
    either from the --dbname option or the configuration file (see option
    ``--config-file`` in ``swh db --help``) in the section named after the
    MODULE argument in most cases.

    When retrieved from within the configuration, the db connection string can
    be looked after from any location in the configuration using the
    <module-or-config-path> option. For example::

        \b
        $ cat conf.yml
        storage:
          cls: pipeline
          steps:
            - cls: masking
              masking_db: postgresql:///?service=swh-masking-proxy
            - cls: buffer
            - cls: postgresql
              db: postgresql://user:passwd@pghost:5433/swh-storage
              objstorage:
                cls: memory

        \b
        $ swh db -C conf.yml init storage  # or
        $ SWH_CONFIG_FILENAME=conf.yml swh db init storage # or
        $ swh db init --dbname postgresql://user:passwd@pghost:5433/swh-storage storage

        $ # to initialize the "main" storage db (expected to be the last element
        $ # of a pipeline config),
        $ # or to initialize the masking_db:
        $ swh db -C conf.yml init -p storage.steps.0

    Usage of --module-config-key is now deprecated in favor of "full-path"
    module/config entry.

    """

    # TODO: sanity check all the incompatible options...
    # XXX it probably does not make much sense to have a non-None flavor when
    # initializing several db at once... this case should raise an error

    args = handle_cmd_args(
        cfg=ctx.obj["config"],
        module=module,
        do_all=initialize_all,
        dbname=dbname,
        is_path=module_is_path,
    )

    for package, fullmodule, backend_class, dbname, cfg in args:
        initialize_one(package, fullmodule, backend_class, flavor, dbname, cfg)


def initialize_one(package, module, backend_class, flavor, dbname, cfg):
    from swh.core.db.db_utils import (
        get_database_info,
        import_swhmodule,
        populate_database_for_package,
        swh_set_db_version,
    )

    logger.debug("db_init %s flavor=%s dbname=%s", module, flavor, dbname)
    dbmodule = f"{package}:{cfg['cls']}"
    initialized, dbversion, dbflavor = populate_database_for_package(
        dbmodule, dbname, flavor
    )
    if dbversion is not None:
        click.secho(
            "ERROR: the database version has been populated by sql init scripts. "
            "This is now deprecated and should not happen any more"
        )
    else:
        # db version has not been populated by sql init scripts (new style),
        # let's do it; instantiate the data source to retrieve the current
        # (expected) db version
        datastore_factory = getattr(import_swhmodule(module), "get_datastore", None)
        if datastore_factory is None and backend_class is not None:

            def datastore_factory(cls, **cfg):
                return backend_class(**cfg)

        if datastore_factory:
            datastore = datastore_factory(**cfg)
            if not hasattr(datastore, "current_version"):
                logger.warning(
                    "Datastore %s does not declare the " "'current_version' attribute",
                    datastore,
                )
            else:
                code_version = datastore.current_version
                logger.info(
                    "Initializing database version to %s from the %s datastore",
                    code_version,
                    module,
                )
                swh_set_db_version(dbname, code_version, desc="DB initialization")

    dbversion = get_database_info(dbname)[1]
    if dbversion is None:
        click.secho(
            "ERROR: database for {} {}{} BUT db version could not be set".format(
                module,
                "initialized" if initialized else "exists",
                f" (flavor {dbflavor})" if dbflavor is not None else "",
            ),
            fg="red",
            bold=True,
        )
    else:
        click.secho(
            "DONE database for {} {}{} at version {}".format(
                module,
                "initialized" if initialized else "exists",
                f" (flavor {dbflavor})" if dbflavor is not None else "",
                dbversion,
            ),
            fg="green",
            bold=True,
        )

    if flavor is not None and dbflavor != flavor:
        click.secho(
            f"WARNING requested flavor '{flavor}' != recorded flavor '{dbflavor}'",
            fg="red",
            bold=True,
        )


@db.command(name="shell", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name or connection URI.",
    default=None,
    show_default=False,
)
@click.option("--module-config-key", help="Module config key to lookup.", default=None)
@click.pass_context
def db_shell(ctx, module, dbname, module_config_key):
    """A subcommand to ease starting a psql shell using swh module configuration file.
    This may be useful for extra troubleshooting session when the other 'swh db' clis
    are not enough.

    """

    from subprocess import run

    if dbname is None:
        # use the db cnx from the config file; the expected config entry is either the given
        # module_config_key or defaulting to the module name (if module_config_key is not
        # provided)
        cfg = ctx.obj["config"].get(module_config_key or module, {})
        dbname, cfg = get_dburl_from_config(cfg)

    if not dbname:
        raise click.BadParameter(
            "Missing the postgresql connection configuration. Either fix your "
            "configuration file or use the [-d|--dbname|--db-name] option."
        )

    dbname_censored = " ".join(
        elem for elem in dbname.split() if not elem.startswith("password=")
    )
    logger.info("Opening database shell for %r", dbname_censored)

    run(["psql", dbname])


@db.command(name="version", context_settings=CONTEXT_SETTINGS)
@click.argument("module", metavar="MODULE-OR-CONFIG-PATH", required=True)
@click.option(
    "--history",
    "show_history",
    help="Show version history.",
    default=False,
    is_flag=True,
)
@click.option(
    "-a",
    "--all",
    "all_backends",
    help="show version for all db found in the config file for the swh 'module'",
    default=False,
    is_flag=True,
)
@click.option("--module-config-key", help="Module config key to lookup.", default=None)
@click.option(
    "-p",
    "--module-is-path",
    "module_is_path",
    help="If set, inpterpret the given 'module' as a config 'json path' within the config file",
    default=False,
    is_flag=True,
)
@click.pass_context
def db_version(
    ctx, module, show_history, all_backends, module_config_key, module_is_path
):
    """Print the database version for the Software Heritage.

    Example::

        \b
        swh db version -d swh-test
        swh db version scheduler
        swh db version scrubber:scrubber_db
        swh db version --all scrubber

    """
    from swh.core.db.db_utils import (
        get_database_info,
        import_swhmodule,
        swh_db_versions,
    )

    backends = handle_cmd_args(
        cfg=ctx.obj["config"],
        module=module,
        do_all=all_backends,
        config_key=module_config_key,
        is_path=module_is_path,
    )

    for package, _, backend_class, cnxstr, cfg in backends:
        db_module, db_version, db_flavor = get_database_info(cnxstr)
        if db_module is None:
            click.secho(
                "WARNING the database does not have a dbmodule table.",
                fg="red",
                bold=True,
            )
            db_module = f"{package}:{cfg['cls']}"

        click.echo("")
        click.secho(f"module: {db_module}", fg="green", bold=True)
        if ":" not in db_module:
            click.secho(
                f"The module registered in the database ({db_module}) needs to be updated;"
                "\nYou should run 'swh db upgrade'",
                fg="yellow",
                bold=True,
            )
        if db_flavor is not None:
            click.secho(f"flavor: {db_flavor}", fg="green", bold=True)

        # instantiate the data source to retrieve the current (expected) db version
        datastore_factory = getattr(import_swhmodule(db_module), "get_datastore", None)
        if not datastore_factory and backend_class is not None:

            def datastore_factory(cls, **cfg):
                return backend_class(**cfg)

        if datastore_factory:
            datastore = datastore_factory(**cfg)
            code_version = datastore.current_version
            click.secho(
                f"current code version: {code_version}",
                fg="green" if code_version == db_version else "red",
                bold=True,
            )

        if not show_history:
            click.secho(f"version: {db_version}", fg="green", bold=True)
        else:
            versions = swh_db_versions(cnxstr)
            for version, tstamp, desc in versions:
                click.echo(f"{version} [{tstamp}] {desc}")


@db.command(name="upgrade", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name or connection URI.",
    default=None,
    show_default=False,
)
@click.option(
    "--to-version",
    type=int,
    help="Upgrade up to version VERSION",
    metavar="VERSION",
    default=None,
)
@click.option(
    "--interactive/--non-interactive",
    help="Do not ask questions (use default answer to all questions)",
    default=True,
)
@click.option(
    "--module-config-key", help="Module configuration key to lookup.", default=None
)
@click.option(
    "-a",
    "--all",
    "upgrade_all",
    help="upgrade all db found in the config file for the swh 'module'",
    default=False,
    is_flag=True,
)
@click.option(
    "-p",
    "--module-is-path",
    "module_is_path",
    help="If set, inpterpret the given 'module' as a config 'json path' within the config file",
    default=False,
    is_flag=True,
)
@click.pass_context
def db_upgrade(
    ctx,
    module,
    dbname,
    to_version,
    interactive,
    module_config_key,
    upgrade_all,
    module_is_path,
):
    """Upgrade the database for given module (to a given version if specified).

    Examples::

        \b
        swh db upgrade storage
        swh db upgrade scheduler --to-version=10
        swh db upgrade scrubber:scrubber_db --to-version=10

    """
    from swh.core.db.db_utils import (
        get_database_info,
        import_swhmodule,
        swh_db_upgrade,
        swh_set_db_module,
    )

    # TODO: mark --module-config-key as deprecated
    # TODO: check options consistency

    backends = handle_cmd_args(
        cfg=ctx.obj["config"],
        module=module,
        do_all=upgrade_all,
        dbname=dbname,
        config_key=module_config_key,
        is_path=module_is_path,
    )

    for package, fullmodule, backend_class, dbname, cfg in backends:
        logger.debug("db_version dbname=%s", dbname)
        go_to_version = to_version
        db_module, db_version, db_flavor = get_database_info(dbname)
        backend = f"{package}:{cfg['cls']}"
        if db_module is None:
            click.secho(
                "Warning: the database does not have a dbmodule table.",
                fg="yellow",
                bold=True,
            )
            if interactive and not click.confirm(
                f"Write the module information ({backend}) in the database?",
                default=True,
            ):
                raise click.BadParameter("Migration aborted.")
        if db_module is None or (db_module != backend and ":" not in db_module):
            # module stored in the db needs updating
            swh_set_db_module(dbname, backend)
            click.secho(
                "The module registered in the database has been updated "
                f"from '{db_module}' to '{backend}'",
                fg="red",
                bold=True,
            )
            db_module, db_version, db_flavor = get_database_info(dbname)

        if db_module != backend:
            raise click.BadParameter(
                f"Error: the given module ({module}) does not match the value "
                f"stored in the database ({db_module})."
            )

        # instantiate the data source to retrieve the current (expected) db version
        datastore_factory = getattr(import_swhmodule(fullmodule), "get_datastore", None)

        if datastore_factory is None and backend_class is not None:

            def datastore_factory(cls, **cfg):
                return backend_class(**cfg)

        if not datastore_factory:
            raise click.UsageError(
                "You cannot use this command on old-style datastore backend {db_module}"
            )
        datastore = datastore_factory(**cfg)
        ds_version = datastore.current_version
        if go_to_version is None:
            go_to_version = ds_version
        if go_to_version > ds_version:
            raise click.UsageError(
                f"The target version {go_to_version} is larger than the current version "
                f"{ds_version} of the datastore backend {db_module}"
            )

        if go_to_version == db_version:
            click.secho(
                f"No migration needed for '{backend}': the current version is {db_version}",
                fg="yellow",
            )
        else:
            new_db_version = swh_db_upgrade(dbname, backend, go_to_version)
            click.secho(f"Migration to version {new_db_version} done", fg="green")
            if new_db_version < ds_version:
                click.secho(
                    "Warning: migration was not complete: "
                    f"the current version is {ds_version}",
                    fg="yellow",
                )


def get_dburl_from_config(cfg):
    if cfg["cls"] == "pipeline":
        # We know the database itself will always
        # come last in a pipeline configuration.
        cfg = cfg["steps"][-1]
    if cfg.get("cls") != "postgresql":
        raise click.BadParameter(
            "Configuration cls must be set to 'postgresql' for this command."
        )
    if "args" in cfg:
        # for bw compat
        cfg = cfg["args"]
    return cfg.get("db"), cfg


def get_dburl_from_config_key(cfg, path):
    # the first section level in the config may contain dotted keys, so deal
    # with it

    for key in cfg:
        if path.startswith(f"{key}."):
            swhmod = key
            cfgpath = path[len(key) + 1 :].split(".")
            break
        if key == path:
            swhmod = key
            cfgpath = []
            break
    cfg = cfg[swhmod]

    for key_e in cfgpath:
        if isinstance(cfg, list):
            cfg = cfg[int(key_e)]
        else:
            cfg = cfg[key_e]

    assert isinstance(cfg, dict)
    if "db" in cfg:
        cnxstr = cfg["db"]
    else:
        # TODO: kill this when possible
        for key in cfg:
            if key.endswith("_db"):
                cnxstr = cfg[key]
                break
        else:
            raise ValueError(
                f"no database connection string found in the configuration at {path}"
            )

    return swhmod, cfg, cnxstr


def handle_cmd_args(
    cfg: Dict[str, Any],
    module: str,
    is_path: bool = False,
    do_all: bool = False,
    dbname: Optional[str] = None,
    config_key: Optional[str] = None,
) -> List[Tuple[str, str, Optional[type], str, Dict[str, Any]]]:
    """Helper function to build the list of backends to handle in a cli command

    For each identified backend, returns a tuple:
      (package, module, backend_class, cnxstr, cfg)
    where:
      - `package`: the (swh) package this backend is implemented in (e.g.
        'storage', 'scheduler' etc.)
      - `module`: the (full path) module in which the backend class can be found
        (e.g. 'swh.storage.postgresql.storage')
      - `backend_class`: the class implementing the backend,
      - `cnxstr`: the (libpq) connection string to the database,
      - `cfg`: the config structure in which this backend is configured; if the
        backend is defined in a nested configuration, this is only the most
        specific config structure for this backend.

    For each backend, the code implementing the backend class needs to be
    registered in the `swh.<package>.classes` entry point under the
    `cfg["cls"]` name. However, there is bw compatibility for swh packages not
    yet updated to register these backends in the entry points.

    If `dbname` is given, return only one element in the list, with a config made
    of {'cls': 'postgresql', 'db': dbname}.

    If `do_all` is True, look for every backend in the configuration (cfg) under
    the section `module`.

    If `module` is a simple word ('storage', 'scheduler', etc.), look for the
    last db backend found in the config file under the `module` section.

    If `is_path` is True, interpret the 'module' as the path to the config entry
    to use in the config file.

    If `module` is an actual module path (e.g. 'storage.proxies.masking'), then
    `dbname` must be given and the configuration is not looked for in the
    config file.

    Note: this rather complex logic will be simplified when all the swh
    packages are updated and do not need bw compat handling code any more.
    """
    from swh.core.config import (
        get_swh_backend_from_fullmodule,
        get_swh_backend_module,
        list_db_config_entries,
    )

    if is_path:
        if do_all:
            raise ValueError("Cannot use both 'all' and a specific config target")
        if dbname:
            raise ValueError("Cannot use both 'dbaname' and a specific config target")

    package = module
    backends = []

    if do_all:
        for cfgmod, path, dbcfg, cnxstr in list_db_config_entries(cfg):
            if cfgmod == module:
                fullmodule, backend_class = get_swh_backend_module(
                    swh_package=cfgmod, cls=dbcfg["cls"]
                )
                backends.append((cfgmod, fullmodule, backend_class, cnxstr, dbcfg))
    else:
        if dbname is not None:
            # default behavior
            if ":" in module:
                module, cls = module.split(":", 1)
            else:
                assert module is not None
                backend_package, backend_cls = get_swh_backend_from_fullmodule(module)
                if backend_package is None:
                    cls = "postgresql"
                else:
                    assert backend_cls is not None
                    module = backend_package
                    cls = backend_cls
            dbcfg = {"cls": cls, "db": dbname}
            fullmodule, backend_class = get_swh_backend_module(
                swh_package=module, cls=cls
            )
            package = module
        else:
            if is_path:
                # read the db access for module 'module' from the config file
                package, dbcfg, dbname = get_dburl_from_config_key(cfg, module)
                # the actual module is retrieved from the entry_point for the cls
                fullmodule, backend_class = get_swh_backend_module(
                    swh_package=package, cls=dbcfg["cls"]
                )
            else:
                # use the db cnx from the config file; the expected config entry is the
                # given module name
                dbname, dbcfg = get_dburl_from_config(cfg.get(config_key or module, {}))
                # the actual module is retrieved from the entry_point for the cls
                fullmodule, backend_class = get_swh_backend_module(
                    swh_package=module, cls=dbcfg["cls"]
                )
        if not dbname:
            raise click.BadParameter(
                "Missing the postgresql connection configuration. Either fix your "
                "configuration file or use the --dbname option."
            )

        backends.append((package, fullmodule, backend_class, dbname, dbcfg))
    return backends



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from dataclasses import dataclass, field
import itertools
from typing import Callable, Generic, Iterable, List, Optional, TypeVar

TResult = TypeVar("TResult")
TToken = TypeVar("TToken")


@dataclass(eq=True)
class PagedResult(Generic[TResult, TToken]):
    """Represents a page of results; with a token to get the next page"""

    results: List[TResult] = field(default_factory=list)
    next_page_token: Optional[TToken] = field(default=None)


def _stream_results(f, *args, page_token, **kwargs):
    """Helper for stream_results() and stream_results_optional()"""
    while True:
        page_result = f(*args, page_token=page_token, **kwargs)
        yield from page_result.results
        page_token = page_result.next_page_token
        if page_token is None:
            break


def stream_results(
    f: Callable[..., PagedResult[TResult, TToken]], *args, **kwargs
) -> Iterable[TResult]:
    """Consume the paginated result and stream the page results"""
    if "page_token" in kwargs:
        raise TypeError('stream_results has no argument "page_token".')
    yield from _stream_results(f, *args, page_token=None, **kwargs)


def stream_results_optional(
    f: Callable[..., Optional[PagedResult[TResult, TToken]]], *args, **kwargs
) -> Optional[Iterable[TResult]]:
    """Like stream_results(), but for functions ``f`` that return an Optional."""
    if "page_token" in kwargs:
        raise TypeError('stream_results_optional has no argument "page_token".')
    res = f(*args, page_token=None, **kwargs)
    if res is None:
        return None
    else:
        if res.next_page_token is None:
            return iter(res.results)
        else:
            return itertools.chain(
                res.results,
                _stream_results(f, *args, page_token=res.next_page_token, **kwargs),
            )



# Copyright (C) 2019-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import warnings

from backports.entry_points_selectable import entry_points
import click

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

logger = logging.getLogger(__name__)


class SWHHelpFormatter(click.HelpFormatter):
    """A subclass of click's HelpFormatter which converts some :ref: to links
    to the SWH documentation
    """

    _template = (
        "{key} key: https://docs.softwareheritage.org/devel/configuration.html#{ref}"
    )

    def write_text(self, text):
        import re

        text = re.sub(
            ":ref:`(?P<ref>cli-config-(?P<key>.*?))`",
            lambda m: self._template.format(**m.groupdict()),
            text,
        )
        text = re.sub(
            ":mod:`(.*?)`",
            r"https://docs.softwareheritage.org/devel/apidoc/\1.html",
            text,
        )
        return super().write_text(text)


class AliasedGroup(click.Group):
    """A simple Group that supports:

    * command aliases
    * notes related to options

    """

    def __init__(self, name=None, commands=None, **attrs):
        self.option_notes = attrs.pop("option_notes", None)
        self.aliases = {}
        super().__init__(name, commands, **attrs)

    def get_command(self, ctx, cmd_name):
        return super().get_command(ctx, self.aliases.get(cmd_name, cmd_name))

    def add_alias(self, name, alias):
        if not isinstance(name, str):
            name = name.name
        self.aliases[alias] = name

    def format_options(self, ctx, formatter):
        click.Command.format_options(self, ctx, formatter)
        if self.option_notes:
            with formatter.section("Notes"):
                formatter.write_text(self.option_notes)
        self.format_commands(ctx, formatter)


def clean_exit_on_signal(signal, frame):
    """Raise a SystemExit exception to let command-line clients wind themselves
    down on exit"""
    raise SystemExit(0)


def validate_loglevel_params(ctx, param, values):
    """Validate the --log-level parameters, with multiple values."""
    if values is None:
        return None

    from swh.core.logging import BadLogLevel, validate_loglevel

    try:
        module_log_levels = [validate_loglevel(value) for value in values]
    except BadLogLevel as e:
        raise click.BadParameter(e)
    return module_log_levels


@click.group(
    context_settings=CONTEXT_SETTINGS,
    cls=AliasedGroup,
    option_notes="""\
If both options are present, --log-level values will override the configuration
in --log-config.

The --log-config YAML must conform to the logging.config.dictConfig schema
documented at https://docs.python.org/3/library/logging.config.html.
""",
)
@click.option(
    "--log-level",
    "-l",
    "log_levels",
    envvar="SWH_LOG_LEVEL",
    default=None,
    callback=validate_loglevel_params,
    multiple=True,
    help=(
        "Log level (defaults to INFO). "
        "Can override the log level for a specific module, by using the "
        "``specific.module:LOGLEVEL`` syntax (e.g. ``--log-level swh.core:DEBUG`` "
        "will enable DEBUG logging for swh.core)."
    ),
)
@click.option(
    "--log-config",
    default=None,
    type=click.Path(exists=True, readable=True),
    help="Python yaml logging configuration file.",
)
@click.option(
    "--sentry-dsn", default=None, help="DSN of the Sentry instance to report to"
)
@click.option(
    "--sentry-debug/--no-sentry-debug",
    default=False,
    hidden=True,
    help="Enable debugging of sentry",
)
@click.pass_context
def swh(ctx, log_levels, log_config, sentry_dsn, sentry_debug):
    """Command line interface for Software Heritage."""
    import signal

    from swh.core.logging import logging_configure

    from ..sentry import init_sentry

    signal.signal(signal.SIGTERM, clean_exit_on_signal)
    signal.signal(signal.SIGINT, clean_exit_on_signal)

    init_sentry(sentry_dsn=sentry_dsn, debug=sentry_debug)

    set_default_loglevel = logging_configure(log_levels, log_config)

    ctx.ensure_object(dict)
    ctx.obj["log_level"] = set_default_loglevel
    ctx.__class__.formatter_class = SWHHelpFormatter


def main():
    # Even though swh() sets up logging, we need an earlier basic logging setup
    # for the next few logging statements
    logging.basicConfig()
    # load plugins that define cli sub commands
    for entry_point in entry_points(group="swh.cli.subcommands"):
        try:
            cmd = entry_point.load()
            if isinstance(cmd, click.BaseCommand):
                # for BW compat, auto add click commands
                warnings.warn(
                    f"{entry_point.name}: automagic addition of click commands "
                    f"to the main swh group is deprecated",
                    DeprecationWarning,
                )
                swh.add_command(cmd, name=entry_point.name)
            # otherwise it's expected to be a module which has been loaded
            # it's the responsibility of the click commands/groups in this
            # module to transitively have the main swh group as parent.
        except Exception as e:
            logger.warning("Could not load subcommand %s: %r", entry_point.name, e)

    return swh(auto_envvar_prefix="SWH")


if __name__ == "__main__":
    main()
